from __future__ import annotations

import logging

from vfbLib.helpers import binaryToIntList
from vfbLib.parsers import BaseParser


logger = logging.getLogger(__name__)


class ExportOptionsParser(BaseParser):
    """
    A parser that reads data a bit field representing export options.
    """

    @classmethod
    def _parse(cls):
        names = {
            0: "use_custom_opentype_export_options",
            # 1: "use_default_opentype_export_options",
            2: "use_custom_cmap_encoding",
        }
        val = cls.read_uint16()
        bits = binaryToIntList(val)
        options = [names.get(i, str(i)) for i in bits]
        return options



from __future__ import annotations

from io import BytesIO
from math import radians, tan
from struct import pack
from typing import Any, List, Tuple
from vfbLib import GLYPH_CONSTANT
from vfbLib.compilers import BaseCompiler
from vfbLib.truetype import TT_COMMAND_CONSTANTS, TT_COMMANDS

import logging


logger = logging.getLogger(__name__)


node_types = {
    "move": 0,
    "line": 1,
    "curve": 3,
    "qcurve": 4,
}


class GlyphCompiler(BaseCompiler):
    @classmethod
    def merge(cls, masters_data: List[Any], data: Any) -> None:
        num_masters = len(masters_data)
        if num_masters < 2:
            return

        for m in range(1, num_masters):
            master_data = masters_data[m]
            # See if there is any data to merge
            if "nodes" in master_data:
                assert "nodes" in data
                for i, tgt in enumerate(data["nodes"]):
                    src = master_data["nodes"][i]
                    for key in ("type", "flags"):
                        assert src[key] == tgt[key]
                    tgt["points"][m] = src["points"][m]

            if "components" in master_data:
                assert "components" in data
                for i, tgt in enumerate(data["components"]):
                    src = master_data["components"][i]
                    for key in ("gid",):
                        assert src[key] == tgt[key]
                    tgt["offsetX"][m] = src["offsetX"][m]
                    tgt["offsetY"][m] = src["offsetY"][m]
                    tgt["scaleX"][m] = src["scaleX"][m]
                    tgt["scaleY"][m] = src["scaleY"][m]

    def _compile_binary(self, data):
        # Imported binary data 8-)
        if not (imported := data.get("imported")):
            return

        logger.warning("Compiling imported binary data is not supported.")
        return

        self.write_uint1(9)

    def _compile_components(self, data):
        # Components
        if not (components := data.get("components")):
            return

        self.write_uint1(5)
        self.write_encoded_value(len(components))
        for component in components:
            self.write_encoded_value(component["gid"])
            for i in range(self.num_masters):
                self.write_encoded_value(component["offsetX"][i])
                self.write_encoded_value(component["offsetY"][i])
                self.write_float(component["scaleX"][i])
                self.write_float(component["scaleY"][i])

    def _compile_glyph_name(self, data):
        # Glyph name
        if not (name := data.get("name")):
            return

        glyph_name = name.encode("cp1252")
        glyph_name_length = len(glyph_name)
        self.write_uint1(1)
        self.write_encoded_value(glyph_name_length)
        self.write_bytes(glyph_name)
        logger.debug(f"Compiling glyph '{name}'")

    def _compile_guides(self, data):
        # Guidelines
        # TODO: Reuse for global guides
        if not (guides := data.get("guides")):
            return

        self.write_uint1(4)
        for direction in ("h", "v"):
            direction_guides = guides.get(direction)
            if direction_guides is None:
                self.write_encoded_value(0)
                continue

            self.write_encoded_value(len(direction_guides[0]))  # first master
            for m in range(self.num_masters):
                for guide in direction_guides[m]:
                    pos = guide["pos"]
                    angle = round(tan(radians(guide["angle"])) * 10000)
                    self.write_encoded_value(pos)
                    self.write_encoded_value(angle)

    def _compile_hints(self, data):
        # PostScript hints
        if not (hints := data.get("hints")):
            return

        self.write_uint1(3)
        for direction in ("h", "v"):
            if direction_hints := hints.get(direction):
                self.write_encoded_value(len(direction_hints))
                for mm_hint in direction_hints:
                    for i in range(self.num_masters):
                        hint = mm_hint[i]
                        self.write_encoded_value(hint["pos"])
                        self.write_encoded_value(hint["width"])
            else:
                self.write_encoded_value(0)

        if not (hintmasks := hints.get("hintmasks")):
            self.write_encoded_value(0)
            return

        # FIXME: Implement writing of hintmasks
        # self.write_encoded_value(len(hintmasks))
        self.write_encoded_value(0)
        logger.warning("Compilation of hint masks is not supported.")

    def _compile_instructions(self, data):
        # TrueType instructions
        if not (tth := data.get("tth")):
            return

        self.write_uint1(0x0A)
        instructions = InstructionsCompiler().compile(tth)
        self.write_encoded_value(len(instructions))
        self.stream.write(instructions)

    def _compile_kerning(self, data):
        # Kerning
        if not (kerning := data.get("kerning")):
            return

        self.write_uint1(6)
        self.write_encoded_value(len(kerning))
        for gid, values in kerning.items():
            self.write_encoded_value(gid)
            for value in values:
                self.write_encoded_value(value)

    def _compile_metrics(self, data):
        # Metrics
        if not (metrics := data.get("metrics")):
            return

        self.write_uint1(2)
        for i in range(self.num_masters):
            x, y = metrics[i]
            self.write_encoded_value(x)
            self.write_encoded_value(y)

    def _compile_outlines(self, data):
        # Outlines
        # A minimal outlines structure is always written:
        self.write_uint1(8)
        self.write_encoded_value(self.num_masters)  # Number of masters

        if not (nodes := data.get("nodes")):
            # 0 nodes with 0 values
            self.write_encoded_value(0)
            self.write_encoded_value(0)
            return

        outlines, num_values = OutlinesCompiler().compile(nodes, self.num_masters)
        self.write_encoded_value(num_values)
        self.stream.write(outlines)

    def _compile(self, data: Any) -> None:

        # Constants?
        self.write_bytes(pack("<4B", *GLYPH_CONSTANT))
        self.num_masters = data["num_masters"]

        self._compile_glyph_name(data)
        self._compile_outlines(data)
        self._compile_metrics(data)
        self._compile_hints(data)
        self._compile_guides(data)
        self._compile_components(data)
        self._compile_kerning(data)
        self._compile_binary(data)
        self._compile_instructions(data)
        self.write_uint1(15)  # End of glyph


class InstructionsCompiler(BaseCompiler):
    def _compile(self, data: Any) -> None:
        self.write_encoded_value(len(data))
        for cmd in data:
            command_id = TT_COMMAND_CONSTANTS[cmd["cmd"]]
            self.write_uint1(command_id)
            params = cmd["params"]
            for param_name in TT_COMMANDS[command_id]["params"]:
                self.write_encoded_value(params[param_name])
        for _ in range(3):
            self.write_encoded_value(0)


class OutlinesCompiler(BaseCompiler):
    def compile(self, data: Any, num_masters: int) -> Tuple[bytes, int]:
        self.num_masters = num_masters
        assert not hasattr(self, "stream")
        self.stream = BytesIO()
        num_values = self._compile(data)
        return self.stream.getvalue(), num_values

    def _compile(self, data: Any) -> int:
        self.write_encoded_value(len(data))  # Number of nodes, may be 0
        num_values = 0
        ref_coords = [[0, 0] for _ in range(self.num_masters)]
        for node in data:
            type_flags = node.get("flags", 0) * 16 + node_types[node["type"]]
            self.write_uint1(type_flags)
            num_values += 1
            for j in range(len(node["points"][0])):
                for i in range(self.num_masters):
                    x, y = node["points"][i][j]
                    refx, refy = ref_coords[i]
                    # Coordinates are written relatively to the previous coords
                    self.write_encoded_value(x - refx)
                    self.write_encoded_value(y - refy)
                    num_values += 2
                    ref_coords[i] = [x, y]
        return 2 * num_values



from __future__ import annotations

import logging

from typing import TYPE_CHECKING, Dict, List, Tuple

if TYPE_CHECKING:
    from vfbLib.typing import ClassFlagDict
    from vfbLib.ufo.typing import UfoGroups


logger = logging.getLogger(__name__)


def transform_groups(
    orig_groups: UfoGroups,
    kerning_class_flags: ClassFlagDict,
    glyphOrder: List[str],
    skip_missing_group_glyphs: bool = False,
) -> UfoGroups:
    # Rename kerning groups by applying the side flags and using the key
    # glyph for naming
    FIRST = 2**10
    SECOND = 2**11
    groups: UfoGroups = {}
    for name, glyphs in orig_groups.items():
        if name.startswith("_"):
            # Kerning group

            # Check for missing glyphs
            missing = [n for n in glyphs if n not in glyphOrder]
            num_missing = len(missing)
            num_glyphs = len(glyphs)
            if missing:
                if num_missing == num_glyphs:
                    logger.warning(
                        f"All glyphs in kerning group '{name}' are missing from "
                        f"the font: {', '.join(missing)}"
                    )
                else:
                    logger.warning(
                        f"{num_missing} of {num_glyphs} glyphs in kerning group "
                        f"'{name}' are missing from the font: {', '.join(missing)}"
                    )
                if skip_missing_group_glyphs:
                    glyphs = list(set(glyphs) - set(missing))
                    if not glyphs:
                        logger.warning("Not adding empty kerning group to the UFO.")
                        continue

            if glyphs:
                key_glyph = glyphs[0]  # Keyglyph is used for group name
                if key_glyph in missing:
                    if num_missing != num_glyphs:
                        logger.warning(
                            f"Key glyph '{key_glyph}' for group '{name}' is missing "
                            "from the font."
                        )

                # Sort group glyphs by glyphOrder
                if not missing:
                    glyphs.sort(key=lambda n: glyphOrder.index(n))
            else:
                # The group is empty, use its original name to build the ufo group name
                groups[name] = []
                key_glyph = name

            if name in kerning_class_flags:
                flags = kerning_class_flags[name][0]
            else:
                # Don't add kerning groups without a side flag
                flags = 0
            for side, sidename in ((FIRST, 1), (SECOND, 2)):
                if flags & side:
                    ufoname = f"public.kern{sidename}.{key_glyph}"
                    if ufoname in groups:
                        logger.warning(f"Duplicate kern{sidename} group: {ufoname}")
                    else:
                        groups[ufoname] = glyphs
        else:
            # Pass non-kerning groups verbatim
            groups[name] = glyphs
    return groups


def build_glyph_to_group_maps(
    groups: UfoGroups,
) -> Tuple[UfoGroups, Dict[str, str], Dict[str, str]]:
    glyph_group_1: Dict[str, str] = {}
    glyph_group_2: Dict[str, str] = {}
    for group in groups.items():
        name, glyphs = group
        if name.startswith("public.kern1"):
            for g in glyphs:
                if g in glyph_group_1:
                    logger.error(
                        f"Multiple kerning groups (kern1) for glyph {g}: "
                        f"{glyph_group_1[g]}, {name}. New group ignored."
                    )
                    continue
                glyph_group_1[g] = name
        elif name.startswith("public.kern2"):
            for g in glyphs:
                if g in glyph_group_2:
                    logger.error(
                        f"Multiple kerning groups (kern2) for glyph {g}: "
                        f"{glyph_group_2[g]}, {name}. New group ignored."
                    )
                    continue
                glyph_group_2[g] = name
    # Rebuild groups to get rid of duplicate kerning groups per glyph and side.
    groups = rebuild_kerning_groups(groups, glyph_group_1, glyph_group_2)
    return groups, glyph_group_1, glyph_group_2


def _build_groups(glyph_to_group_mapping: Dict[str, str]) -> UfoGroups:
    groups: UfoGroups = {}
    for name, group in glyph_to_group_mapping.items():
        if group in groups:
            groups[group].append(name)
        else:
            groups[group] = [name]
    return groups


def rebuild_kerning_groups(
    groups: UfoGroups, glyph_group_1: Dict[str, str], glyph_group_2: Dict[str, str]
) -> UfoGroups:
    """
    Rebuild the kerning groups from the left and right side glyph to group mappings.
    This should get rid of any duplicate groups per glyph.
    """
    # Filter groups so only non-kerning groups are kept
    new_groups: UfoGroups = {
        n: g for n, g in groups.items() if not n.startswith("public.kern")
    }
    new_groups.update(_build_groups(glyph_group_1))
    new_groups.update(_build_groups(glyph_group_2))
    return new_groups



from __future__ import annotations

from io import BufferedReader, BytesIO
from struct import unpack


def read_encoded_value(stream: BufferedReader | BytesIO, signed=True) -> int:
    val = int.from_bytes(stream.read(1), byteorder="little")
    if val == 0:
        raise EOFError

    if val < 0x20:
        raise ValueError

    elif val < 0xF7:
        # -107 to 107, represented by 1 byte
        return val - 0x8B

    elif val <= 0xFA:
        # 108 to 1131, represented by 2 bytes
        val2 = int.from_bytes(stream.read(1), byteorder="little")
        # val - 0x8B + (val - 0xF7) * 0xFF + val2
        return 0x100 * val - 0xF694 + val2

    elif val <= 0xFE:
        # -108 to -1131, represented by 2 bytes
        val2 = int.from_bytes(stream.read(1), byteorder="little")
        # 0x8F - val - (val - 0xFB) * 0xFF - val2
        return -0x100 * val + 0xFA94 - val2

    elif val == 0xFF:
        # 4-byte integer follows
        decoded = int.from_bytes(stream.read(4), byteorder="big", signed=signed)
        return decoded

    raise ValueError


def read_doubles(num, stream):
    # Read a number of doubles from the stream and return them
    return unpack(num * "d", stream.read(num * 8))


def read_floats(num, stream):
    # Read a number of floats from the stream and return them
    return unpack(num * "f", stream.read(num * 4))



from __future__ import annotations

import logging

from typing import Dict, List
from vfbLib.parsers import BaseParser, read_encoded_value


logger = logging.getLogger(__name__)


class NameRecordsParser(BaseParser):
    @classmethod
    def _parse(cls):
        stream = cls.stream
        num = read_encoded_value(stream)
        result = []
        for _ in range(num):
            nameID = read_encoded_value(stream)
            platID = read_encoded_value(stream)
            encID = read_encoded_value(stream)
            langID = read_encoded_value(stream)
            name_length = read_encoded_value(stream)
            name_codes = [read_encoded_value(stream) for _ in range(name_length)]
            name = ""
            for c in name_codes:
                try:
                    char = chr(c)
                    # Fix platform-specific encodings for Mac
                    if platID == 1 and encID == 0:
                        char = c.to_bytes().decode("macroman")
                except ValueError:
                    char = "\ufeff"
                name += char
            result.append([nameID, platID, encID, langID, name])

        assert stream.read() == b""
        return result


class OpenTypeClassParser(BaseParser):
    """
    A parser that reads data as strings and returns it formatted to represent an
    OpenType class
    """

    @classmethod
    def _parse(cls) -> Dict[str, List[str] | str]:
        s = cls.stream.read().decode(cls.encoding).strip("\u0000 ")
        if ":" not in s:
            logger.warning(f"Malformed OpenType class: {s}")
            return {"str": s, "err": "PARSE_ERROR"}
            # raise ValueError

        name, contents = s.split(":")

        glyphs = []
        for glyph in contents.split(" "):
            glyph = glyph.strip()
            if glyph:
                glyphs.append(glyph)
        return {"name": name, "glyphs": glyphs}


class OpenTypeStringParser(BaseParser):
    """
    A parser that reads data as a strings and returns it as a list.
    """

    @classmethod
    def _parse(cls) -> List[str]:
        s = cls.stream.read().decode(cls.encoding).strip("\u0000 ")
        return s.splitlines()


class StringParser(BaseParser):
    """
    A parser that reads data as strings.
    """

    @classmethod
    def _parse(cls):
        return cls.stream.read().decode(cls.encoding).strip("\u0000 ")



from __future__ import annotations

import logging

import xml.etree.cElementTree as elementTree

from vfbLib.ufo.typing import UfoHintingV2, UfoHintSet
from vfbLib.ufo.vfb2ufo import PS_GLYPH_LIB_KEY
from typing import TYPE_CHECKING, Dict, List, Tuple

if TYPE_CHECKING:
    from vfbLib.typing import Hint, HintTuple
    from vfbLib.ufo.glyph import VfbToUfoGlyph
    from vfbLib.ufo.paths import UfoMasterGlyph


logger = logging.getLogger(__name__)


def normalize_hint(hint: Tuple[str, int, int]):
    direction, pos, width = hint
    if width < 0:
        if width not in (-21, -20):  # Skip ghost hints
            pos = pos + width
            width = abs(width)
    return (direction, pos, width)


def normalize_hint_dict(hint: Hint, name: str = "dummy"):
    return normalize_hint((name, hint["pos"], hint["width"]))


def build_ps_glyph_hints(
    mmglyph: VfbToUfoGlyph,
    glyph: UfoMasterGlyph,
    master_hints: Dict[str, List[str | HintTuple]],
) -> None:
    # Set the master-specific hints from data to the glyph lib
    # Use the format defined in UFO3, not what FL does.
    # https://github.com/adobe-type-tools/psautohint/blob/master/python/psautohint/ufoFont.py
    # https://unifiedfontobject.org/versions/ufo3/glyphs/glif/#publicpostscripthints
    hint_sets = []
    stems: List[str | HintTuple] = []
    hint_set: UfoHintSet = UfoHintSet(pointTag="0", stems=stems)
    if mmglyph.hintmasks:
        for mask in mmglyph.hintmasks:
            for d in ("h", "v"):
                if d in mask:
                    hint_index = mask[d]
                    hint: HintTuple | str = master_hints[d][hint_index]
                    hint_set["stems"].append(hint)
            if "r" in mask:
                hint_set["pointTag"] = mmglyph.get_point_label(
                    index=int(hint_set["pointTag"]),
                    code="PSHintReplacement",
                    start_count=0,
                )
                hint_sets.append(hint_set)
                node_index = mask["r"]
                # FIXME: What do negative values mean?
                if node_index < 0:
                    node_index = abs(node_index) - 1
                stems = []
                hint_set = UfoHintSet(pointTag=str(node_index), stems=stems)

        if hint_set["stems"]:
            # Append the last hint set
            hint_set["pointTag"] = mmglyph.get_point_label(
                index=int(hint_set["pointTag"]),
                code="PSHintReplacement",
                start_count=0,
            )
            hint_sets.append(hint_set)
    else:
        # Only one hint set, always make a hint set with first point
        for d in ("h", "v"):
            for hint in master_hints[d]:
                hint_set["stems"].append(hint)
        if hint_set["stems"]:
            hint_set["pointTag"] = mmglyph.get_point_label(
                index=int(hint_set["pointTag"]),
                code="PSHintReplacement",
                start_count=0,
            )
            hint_sets = [hint_set]

    # Reformat stems from sortable tuples to str required by UFO spec
    for hint_set in hint_sets:
        hint_set["stems"] = [
            f"{h[0]} {h[1]} {h[2]}"
            for h in sorted(set(hint_set["stems"]))
            # if isinstance(h, tuple)
        ]

    if hint_sets:
        if not hasattr(glyph, "lib"):
            glyph.lib = {}
        glyph.lib[PS_GLYPH_LIB_KEY] = {
            # "id": "FIXME",
            "hintSetList": hint_sets,
            # "flexList": [],
        }


def get_master_hints(
    mmglyph: VfbToUfoGlyph, master_index=0
) -> Dict[str, List[str | HintTuple]]:
    hints: Dict[str, List[str | HintTuple]] = {"h": [], "v": []}

    # Hints
    for d in "hv":
        dh = mmglyph.mm_hints[d]
        for mm_hints in dh:
            hint = mm_hints[master_index]
            hint = normalize_hint_dict(hint, f"{d}stem")
            hints[d].append(hint)

    # Links
    if not mmglyph.links:
        return hints

    # Convert links to hints
    for i, axis in enumerate("xy"):
        direction_links = mmglyph.links[axis]
        for link in direction_links:
            isrc, itgt = link  # indices of source and target node
            src = mmglyph.mm_nodes[isrc]
            src_pos = src["points"][master_index][0][i]
            pos = src_pos
            if itgt == -1:  # Bottom ghost
                width = -21
                pos = src_pos - width
            elif itgt == -2:  # Top ghost
                width = -20
            else:
                tgt = mmglyph.mm_nodes[itgt]
                tgt_pos = tgt["points"][master_index][0][i]
                width = tgt_pos - src_pos
                # pos = min(src_pos, tgt_pos)

            d = "v" if axis == "x" else "h"
            # Don't normalize those values, the above code already did that
            hint = (f"{d}stem", pos, width)
            # hint = normalize_hint((f"{d}stem", pos, width))
            hints[d].append(hint)

    return hints


def update_adobe_hinting(data) -> UfoHintingV2:
    # Convert Adobe hinting data from v1 to v2.
    # https://github.com/adobe-type-tools/psautohint/blob/master/python/psautohint/ufoFont.py
    try:
        # Data may be base64-encoded
        data = data.decode()
    except AttributeError:
        pass
    if not isinstance(data, str):
        # V1 data is stored as str, so if it is not a str, we have nothing to do
        return data

    v2: UfoHintingV2 = {
        # "flexList": [],
        # "id": "",
    }
    root = elementTree.fromstring(data)
    hintset: UfoHintSet | None = None
    hintSetList: List[UfoHintSet] = []
    for el in root.iter():
        if el.tag == "hintSetList":
            hintSetList = []
        elif el.tag == "hintset":
            if hintset:
                hintSetList.append(hintset)
            hintset = {
                "pointTag": el.attrib["pointTag"],
                "stems": [],
            }
        elif el.tag in ("hstem", "vstem"):
            if hintset is not None:
                tag, pos, width = normalize_hint(
                    (el.tag, int(el.attrib["pos"]), int(el.attrib["width"]))
                )
                hintset["stems"].append(f"{tag} {pos} {width}")
    if hintset:
        hintSetList.append(hintset)
    if hintSetList:
        v2["hintSetList"] = hintSetList

    # Remove empty entries
    empty = [k for k in v2 if not v2[k]]
    for k in empty:
        del v2[k]
    return v2



from __future__ import annotations

import logging

from base64 import b64encode
from copy import deepcopy
from typing import Any, Dict, List, TYPE_CHECKING
from vfbLib.ufo.guides import apply_guide_properties, get_master_guides
from vfbLib.ufo.pshints import build_ps_glyph_hints, get_master_hints
from vfbLib.ufo.tth import set_tth_lib
from vfbLib.ufo.vfb2ufo import TT_GLYPH_LIB_KEY

if TYPE_CHECKING:
    from fontTools.pens.pointPen import AbstractPointPen
    from vfbLib.typing import Anchor
    from vfbLib.ufo.glyph import VfbToUfoGlyph
    from vfbLib.ufo.typing import UfoComponent, UfoContour


logger = logging.getLogger(__name__)


class UfoMasterGlyph:
    def __init__(
        self,
        mm_glyph: VfbToUfoGlyph,
        glyph_order: List[str],
        master_index: int,
    ) -> None:
        self.mm_glyph = mm_glyph
        self.glyph_order = glyph_order
        self.master_index = master_index

        self.lib: Dict[str, Any] = {}
        self.anchors: List[Anchor] = []
        self.guidelines: List = []
        self.unicodes: List[int] = []
        self.width: int = 0
        self.height: int = 0

        self.components: List[UfoComponent] = []
        self.contours: List[UfoContour] = []
        self.rename_points: Dict[str, str] = {}
        self.tth_commands: List[Dict[str, str | bool]] = []

    @property
    def name(self) -> str | None:
        return self.mm_glyph.name

    def build(
        self,
        minimal: bool = False,
        include_ps_hints: bool = True,
        encode_data_base64: bool = False,
    ) -> None:
        # Extract the single master glyph from an mm glyph. The main method.
        # Copy glyph lib
        self.lib = deepcopy(self.mm_glyph.lib)
        self.tth_commands = deepcopy(self.mm_glyph.tth_commands)
        if include_ps_hints:
            self._extract_master_ps_hints()
        self._extract_master_tt_hints()
        self._extract_master_contours()
        self._finalize_point_labels(include_ps_hints)
        self._extract_master_anchors()
        if not minimal:
            self._extract_master_guides()
        self._finalize_lib(encode_data_base64)
        self.unicodes = self.mm_glyph.unicodes.copy()
        self.width, self.height = self.mm_glyph.mm_metrics[self.master_index]

    def drawPoints(self, pen: AbstractPointPen) -> None:
        """
        Draw the glyph onto the supplied point pen.
        """
        for contour in self.contours:
            pen.beginPath()
            for segment_type, smooth, name, pt in contour:
                pen.addPoint(pt, segment_type, name=name, smooth=smooth)
            pen.endPath()

        for gn, tr in self.components:
            pen.addComponent(gn, tr)

    def _extract_master_anchors(self) -> None:
        if self.mm_glyph.anchors is None:
            return

        # Copy anchors from the mm glyph
        self.anchors = deepcopy(self.mm_glyph.anchors)

        if self.mm_glyph.mm_anchors is None:
            return

        # Apply master anchor positions
        for j, anchor in enumerate(self.mm_glyph.mm_anchors):
            self.anchors[j]["x"] = anchor["x"][self.master_index]
            self.anchors[j]["y"] = anchor["y"][self.master_index]

    def _extract_master_contours(self) -> None:
        """
        Extract the contours and components from the mm contours.
        """
        self.contours = []
        self.rename_points = {}
        path_is_open = False
        in_qcurve = False
        if hasattr(self.mm_glyph, "mm_nodes"):
            contour: UfoContour = []
            for i, n in enumerate(self.mm_glyph.mm_nodes):
                name: str | None = self.mm_glyph.point_labels.get(i, None)
                nodes = n["points"][self.master_index]
                pt = nodes[0]
                segment_type = n["type"]
                flags = n["flags"]
                smooth = bool(flags & 1)

                if segment_type == "move":
                    self._append_contour(contour, path_is_open)
                    contour = [("move", smooth, name, pt)]
                    path_is_open = bool(flags & 8)
                    in_qcurve = False

                elif segment_type == "line":
                    if in_qcurve:
                        contour.append(("qcurve", smooth, name, pt))
                        in_qcurve = False
                    else:
                        contour.append(("line", smooth, name, pt))

                elif segment_type == "curve":
                    pt, c1, c2 = nodes
                    contour.append((None, False, None, c1))
                    contour.append((None, False, None, c2))
                    contour.append(("curve", smooth, name, pt))
                    in_qcurve = False

                elif segment_type == "qcurve":
                    contour.append((None, False, name, pt))
                    in_qcurve = True

                else:
                    logger.error(f"Unknown segment type: {segment_type}")
                    raise ValueError

            if contour:
                self._append_contour(contour, path_is_open)

        self.components = []
        if hasattr(self.mm_glyph, "mm_components"):
            for c in self.mm_glyph.mm_components:
                transform = (
                    c["scaleX"][self.master_index],
                    0.0,
                    0.0,
                    c["scaleY"][self.master_index],
                    c["offsetX"][self.master_index],
                    c["offsetY"][self.master_index],
                )
                self.components.append((self.glyph_order[c["gid"]], transform))

    def _extract_master_guides(self) -> None:
        if self.mm_glyph.mm_guides is None:
            return

        master_guides = get_master_guides(self.mm_glyph.mm_guides, self.master_index)
        apply_guide_properties(master_guides, self.mm_glyph.guide_properties)
        if master_guides:
            self.guidelines = master_guides

    def _extract_master_ps_hints(self) -> None:
        """
        Apply master hint positions and widths.
        """
        master_hints = get_master_hints(
            mmglyph=self.mm_glyph, master_index=self.master_index
        )
        build_ps_glyph_hints(
            mmglyph=self.mm_glyph,
            glyph=self,
            master_hints=master_hints,
        )

    def _extract_master_tt_hints(self) -> None:
        if self.mm_glyph.tth_commands is None:
            return

        set_tth_lib(self, self.tth_commands)

    def _finalize_lib(self, encode_data_base64: bool = False) -> None:
        # If requested, apply base64 encoding to TT lib
        if encode_data_base64 and TT_GLYPH_LIB_KEY in self.lib:
            data = self.lib[TT_GLYPH_LIB_KEY]
            if not isinstance(data, bytes):
                self.lib[TT_GLYPH_LIB_KEY] = b64encode(data.encode("ascii"))

    def _finalize_point_labels(self, include_ps_hints: bool = True) -> None:
        self._finalize_tt_point_labels()
        if include_ps_hints:
            self._finalize_ps_point_labels()

    def _finalize_ps_point_labels(self) -> None:
        """
        Rename the ps-hinted point labels according to the supplied rename_dict.
        """
        if not self.rename_points:
            return

        # FIXME

    def _finalize_tt_point_labels(self) -> None:
        """
        Rename the tt-hinted point labels according to the supplied rename_dict.
        """
        if not self.rename_points:
            return

        for cmd in self.tth_commands:
            code = cmd["code"]
            if code in (
                "alignb",
                "alignt",
                "alignh",
                "alignv",
                "mdeltah",
                "mdeltav",
                "fdeltah",
                "fdeltav",
            ):
                self._rename_point_in_cmd(cmd, "point")
            elif code in ("singleh", "singlev", "doubleh", "doublev"):
                self._rename_point_in_cmd(cmd, "point1")
                self._rename_point_in_cmd(cmd, "point2")
            elif code in ("interpolateh", "interpolatev"):
                self._rename_point_in_cmd(cmd, "point")
                self._rename_point_in_cmd(cmd, "point1")
                self._rename_point_in_cmd(cmd, "point2")
            else:
                logger.error(f"Unknown TT command: {code}")
                raise ValueError

    def _rename_point_in_cmd(self, cmd, key: str) -> None:
        """
        Rename a point label in a TTH cmd according to the current rename_points dict.
        """
        pt_name = cmd[key]
        if pt_name in self.rename_points:
            cmd[key] = self.rename_points[pt_name]

    def _append_contour(self, contour: UfoContour, path_is_open: bool) -> None:
        """
        Append the contour to the glyph's contours, applying closepath optimizations.
        """
        if contour:
            ct = self._flush_contour(contour, path_is_open)
            self.contours.append(ct)

    def _apply_closepath(self, contour: UfoContour) -> None:
        """
        Apply closepath optimizations to the contour.
        """
        if contour[-1][3] == contour[0][3]:
            # Equal coords, use closepath to draw the last line
            name = contour[0][2]
            contour[0] = contour.pop()
            if name is not None:
                # Keep old point name
                if contour[0][2] is None:
                    t, smooth, _, pt = contour[0]
                    contour[0] = (t, smooth, name, pt)
                else:
                    logger.warning(f"Glyph '{self.name}':")
                    logger.warning(
                        f"Point name conflict in {contour[0]} vs. {name} while "
                        f"applying closepath. Not applying old name ({name})"
                    )
                    # Note the old and new name so the labels can be updated
                    self.rename_points[name] = contour[0][2]

        else:
            _, smooth, name, pt = contour[0]
            contour[0] = ("line", smooth, name, pt)

    def _flush_contour(self, contour: UfoContour, path_is_open: bool) -> UfoContour:
        """
        Post-process a contour before it is appended to the glyph's contours.
        """
        if not path_is_open:
            last_type = contour[-1][0]
            if last_type in ("line", "curve", "qcurve"):
                self._apply_closepath(contour)

            elif last_type is None:
                _, smooth, name, pt = contour[0]
                contour[0] = ("qcurve", smooth, name, pt)
        return contour



from __future__ import annotations

from datetime import datetime


TIME_ZERO = 2082841200


def convert_timestamp(t: int) -> datetime:
    """
    Convert the FontLab timestamp to a datetime object.
    """
    # 1970-01-01 is the earliest timestamp in FL notation, enforce it
    if t < TIME_ZERO:
        t = TIME_ZERO

    return datetime.utcfromtimestamp(t - TIME_ZERO)



from __future__ import annotations

import logging

from fontTools.misc.textTools import hexStr
from typing import TYPE_CHECKING, Any, Dict
from vfbLib.parsers.header import VfbHeaderParser

if TYPE_CHECKING:
    from io import BufferedReader


logger = logging.getLogger(__name__)


class VfbHeader:
    def __init__(self) -> None:
        # The original or compiled binary data
        self.data: bytes | None = None
        # The decompiled data
        self.decompiled = None
        # Has the data been modified, i.e. it needs recompilation
        self.modified = False
        # The parser which can convert data to decompiled
        self.parser = VfbHeaderParser
        # The size of the compiled data
        self.size = 0

    def as_dict(self) -> Dict[str, Any]:
        d = {
            "size": self.size,
            "decompiled": self.decompiled,
            "modified": self.modified,
            "parser": "VfbHeaderParser",
        }
        if self.data:
            d["data"] = hexStr(self.data)
        return d

    def compile(self) -> bytes:
        logger.error("Compiling the VFB header is not supported yet.")

        if self.data:
            self.size = len(self.data)
        else:
            self.size = 0

        self.modified = False

        if self.data is None:
            return b""

        return self.data

    def read(self, stream: BufferedReader) -> None:
        self.decompiled, self.size = self.parser.parse(stream)
        stream.seek(0)
        self.data = stream.read(self.size)



from __future__ import annotations

import logging

from typing import TYPE_CHECKING, List, Literal, Sequence
from math import atan2, degrees
from vfbLib.parsers import BaseParser, read_encoded_value
from vfbLib.typing import Guide, GuideDict, GuideProperty

if TYPE_CHECKING:
    from io import BytesIO


logger = logging.getLogger(__name__)


DIRECTIONS: Sequence[Literal["h", "v"]] = ("h", "v")


def parse_guides(stream: BytesIO, num_masters: int) -> GuideDict:
    # Common parser for glyph and global guides
    guides: GuideDict = {
        "h": [[] for _ in range(num_masters)],
        "v": [[] for _ in range(num_masters)],
    }
    for d in DIRECTIONS:
        num_guides = read_encoded_value(stream)
        for _ in range(num_guides):
            for m in range(num_masters):
                pos = read_encoded_value(stream)
                angle = degrees(atan2(read_encoded_value(stream), 10000))
                guides[d][m].append(Guide(pos=pos, angle=angle))

    return guides


class GlobalGuidesParser(BaseParser):
    @classmethod
    def _parse(cls) -> GuideDict:
        assert cls.master_count is not None
        guides = parse_guides(cls.stream, cls.master_count)
        assert cls.stream.read() == b""
        return guides


class GuidePropertiesParser(BaseParser):
    @classmethod
    def _parse(cls) -> List:
        stream = cls.stream
        guides = []
        for _ in range(2):
            while True:
                index = read_encoded_value(stream)
                if index == 0:
                    break

                g = GuideProperty(index=index)

                color = read_encoded_value(stream)
                if color > -1:
                    g["color"] = "#" + hex(color & ~0xFF00000000)[2:]

                name_length = read_encoded_value(stream)
                if name_length > 0:
                    name = cls.stream.read(name_length).decode("cp1252")
                    g["name"] = name

                guides.append(g)

        assert stream.read() == b""
        return guides



from __future__ import annotations

import logging

from fontTools.pens.pointPen import AbstractPointPen
from typing import TYPE_CHECKING, Any, Dict, Tuple

if TYPE_CHECKING:
    from vfbLib.vfb.glyph import VfbGlyph, VfbGlyphMaster
    from vfbLib.vfb.vfb import Vfb


logger = logging.getLogger(__name__)


class VfbGlyphPointPen(AbstractPointPen):
    # FIXME: Only supports TrueType curves
    def __init__(self, glyph: VfbGlyph | VfbGlyphMaster, glyphSet: Vfb) -> None:
        """A PointPen to draw into the VFB glyph.

        Args:
            glyph (VfbGlyph): The glyph to draw into.
        """
        self.glyph = glyph
        self.glyphSet = glyphSet
        self.currentPath = None
        self.in_qcurve = False
        if self.glyph.master_index == 0:
            self.target = self.glyph.entry.decompiled
        else:
            if self.glyph.entry.temp_masters is None:
                self.glyph.entry.temp_masters = [[] * self.glyphSet.num_masters]
            self.target = self.glyph.entry.temp_masters[self.glyph.master_index]

    def beginPath(self) -> None:
        self.currentPath = []

    def endPath(self) -> None:
        self.target["nodes"].extend(self.currentPath)
        self.currentPath = None

    def addPoint(
        self,
        pt: Tuple[int, int],
        segmentType: str | None = None,
        smooth: bool = False,
        name: str | None = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        assert self.currentPath is not None

        flags = 0
        x, y = pt
        node = {
            "type": None,
            "flags": flags,
            "points": [
                [(round(x), round(y))] for _ in range(self.glyphSet.num_masters)
            ],
        }

        if segmentType == "qcurve":
            self.in_qcurve = True

        if not self.currentPath:
            # Begin a new path

            if segmentType == "move":  # Open path
                flags += 8

            # VFB first node always has type "move"
            node["type"] = "move"

        else:
            # During path
            if segmentType is None:
                node["type"] = "qcurve"
            elif segmentType == "qcurve":
                if self.in_qcurve:
                    node["type"] = "line"
                    self.in_qcurve = False
            elif segmentType == "line":
                node["type"] = "line"
            else:
                print(f"Unsupported segment type: {segmentType}")
                raise ValueError

        node["flags"] = flags
        self.currentPath.append(node)

    def addComponent(
        self,
        baseGlyphName: str,
        transformation: Tuple[float, float, float, float, float, float],
        identifier: str | None = None,
        **kwargs: Any,
    ) -> None:
        assert self.currentPath is None
        base_index = self.glyphSet.glyph_order.index(baseGlyphName)
        if base_index == -1:
            raise (KeyError, f"Base glyph not found: '{baseGlyphName}'")

        xx, xy, yx, yy, dx, dy = transformation

        if "components" not in self.target:
            self.target["components"] = []
        self.target["components"].append(
            {
                "gid": base_index,
                "offsetX": [dx] * self.glyphSet.num_masters,
                "offsetY": [dy] * self.glyphSet.num_masters,
                "scaleX": [xx] * self.glyphSet.num_masters,
                "scaleY": [yy] * self.glyphSet.num_masters,
            }
        )



from __future__ import annotations

from vfbLib.compilers.glyph import GlyphCompiler

from vfbLib.parsers import (
    BaseParser,
    # EncodedValueParser,
    EncodedValueListParser,
    EncodedValueListWithCountParser,
    EncodedKeyValuesParser,
    EncodedKeyValuesParser1742,
    GaspParser,
    GlyphEncodingParser,
    OpenTypeClassFlagsParser,
)
from vfbLib.parsers.bitmap import BackgroundBitmapParser, GlyphBitmapParser
from vfbLib.parsers.cmap import CustomCmapParser
from vfbLib.parsers.glyph import (
    GlyphAnchorsParser,
    GlyphAnchorsSuppParser,
    GlyphGDEFParser,
    GlyphOriginParser,
    GlyphParser,
    GlyphUnicodeParser,
    GlyphUnicodeSuppParser,
    LinkParser,
    MaskParser,
)
from vfbLib.parsers.guides import (
    GlobalGuidesParser,
    GuidePropertiesParser,
)
from vfbLib.parsers.mm import (
    AnisotropicInterpolationsParser,
    AxisMappingsCountParser,
    AxisMappingsParser,
    MasterLocationParser,
    PrimaryInstancesParser,
)
from vfbLib.parsers.numeric import (
    DoubleListParser,
    DoubleParser,
    FloatListParser,
    IntParser,
    IntListParser,
    PanoseParser,
    SignedIntParser,
)
from vfbLib.parsers.options import ExportOptionsParser
from vfbLib.parsers.ps import PostScriptInfoParser
from vfbLib.parsers.text import (
    NameRecordsParser,
    OpenTypeClassParser,
    OpenTypeStringParser,
    StringParser,
)
from vfbLib.parsers.truetype import (
    TrueTypeInfoParser,
    TrueTypeStemPpemsParser,
    TrueTypeStemsParser,
    TrueTypeZoneDeltasParser,
    TrueTypeZonesParser,
)


# fmt: off
parser_classes = {
    513: ("513", BaseParser, None),
    527: ("527", BaseParser, None),
    1024: ("sgn", StringParser, None),
    1025: ("ffn", StringParser, None),
    1026: ("psn", StringParser, None),
    1027: ("tfn", StringParser, None),
    1028: ("weight_name", StringParser, None),
    1029: ("Italic Angle", DoubleParser, None),
    1030: ("underlinePosition", SignedIntParser, None),
    1031: ("underlineThickness", IntParser, None),
    1034: ("Monospaced", IntParser, None),
    1037: ("copyright", StringParser, None),
    1038: ("description", StringParser, None),
    1039: ("manufacturer", StringParser, None),
    1044: ("Type 1 Unique ID", SignedIntParser, None),
    1046: ("version full", StringParser, None),
    1047: ("Slant Angle", DoubleParser, None),
    1048: ("weight", SignedIntParser, None),  # Weight Class
    1054: ("MS Character Set", IntParser, None),
    1056: ("Menu Name", StringParser, None),
    1057: ("PCL ID", IntParser, None),
    1058: ("VP ID", IntParser, None),
    1059: ("1059", BaseParser, None),
    1060: ("MS ID", IntParser, None),
    1061: ("trademark", StringParser, None),
    1062: ("designer", StringParser, None),
    1063: ("designerURL", StringParser, None),
    1064: ("manufacturerURL", StringParser, None),
    1065: ("width_name", StringParser, None),
    1066: ("Default Glyph", StringParser, None),
    1068: ("1068", EncodedValueListWithCountParser, None),
    1069: ("License", StringParser, None),
    1070: ("License URL", StringParser, None),
    1090: ("FOND Family ID", IntParser, None),
    1092: ("FOND Name", StringParser, None),
    1093: ("1093", BaseParser, None),
    1118: ("panose", PanoseParser, None),
    1121: ("vendorID", StringParser, None),
    1127: ("Style Name", StringParser, None),
    1128: ("version", StringParser, None),
    1129: ("UniqueID", StringParser, None),
    1130: ("versionMajor", IntParser, None),
    1131: ("versionMinor", IntParser, None),
    1132: ("year", IntParser, None),
    1133: ("Type 1 XUIDs", IntListParser, None),
    1134: ("Type 1 XUIDs Count", IntParser, None),
    1135: ("upm", IntParser, None),
    1136: ("PCLT Table", EncodedValueListParser, None),
    1137: ("tsn", StringParser, None),
    1138: ("Name Records", NameRecordsParser, None),
    1139: ("OT Mac Name", StringParser, None),
    1140: ("1140", BaseParser, None),
    1141: ("Custom CMAPs", CustomCmapParser, None),
    1247: ("Primary Instance Locations", DoubleListParser, None),
    1250: ("Glyph Unicode", GlyphUnicodeParser, None),
    1253: ("Glyph Unicode Non-BMP", GlyphUnicodeSuppParser, None),
    1254: ("Primary Instances", PrimaryInstancesParser, None),
    1255: ("TrueType Zones", TrueTypeZonesParser, None),
    1264: ("TrueType Info", TrueTypeInfoParser, None),
    1265: ("Gasp Ranges", GaspParser, None),
    1267: ("Selection", IntParser, None),
    1268: ("TrueType Stem PPEMs", TrueTypeStemPpemsParser, None),
    1269: ("TrueType Stems", TrueTypeStemsParser, None),
    1270: ("hhea_line_gap", IntParser, None),
    1271: ("1271", EncodedValueListParser, None),
    1272: ("Pixel Snap", IntParser, None),
    1273: ("TrueType Zone Deltas", TrueTypeZoneDeltasParser, None),
    1274: ("Zone Stop PPEM", IntParser, None),
    1275: ("Code Stop PPEM", IntParser, None),
    1276: ("openTypeFeatures", OpenTypeStringParser, None),
    1277: ("OpenType Class", OpenTypeClassParser, None),
    1278: ("hhea_ascender", SignedIntParser, None),
    1279: ("hhea_descender", SignedIntParser, None),
    1294: ("Global Guides", GlobalGuidesParser, None),
    1296: ("Global Guide Properties", GuidePropertiesParser, None),
    1410: ("1410", BaseParser, None),
    1500: ("Encoding", GlyphEncodingParser, None),
    1501: ("Encoding Mac", GlyphEncodingParser, None),
    1502: ("1502", IntParser, None),
    1503: ("Master Count", IntParser, None),
    1504: ("Master Name", StringParser, None),
    1505: ("Master Location", MasterLocationParser, None),
    1513: ("Axis Count", IntParser, None),
    1514: ("Axis Name", StringParser, None),
    1515: ("Axis Mappings Count", AxisMappingsCountParser, None),
    1516: ("Axis Mappings", AxisMappingsParser, None),
    1517: ("Default Weight Vector", DoubleListParser, None),  # Interp. coeffs for all masters  # noqa: E501
    1523: ("Anisotropic Interpolation Mappings", AnisotropicInterpolationsParser, None),
    1524: ("1524", EncodedValueListParser, None),
    1530: ("Blue Values Count", IntParser, None),
    1531: ("Other Blues Count", IntParser, None),
    1532: ("Family Blues Count", IntParser, None),
    1533: ("Family Other Blues Count", IntParser, None),
    1534: ("StemSnapH Count", IntParser, None),
    1535: ("StemSnapV Count", IntParser, None),
    1536: ("PostScript Info", PostScriptInfoParser, None),
    1604: ("1604", IntParser, None),
    1742: ("1742", EncodedKeyValuesParser1742, None),
    1743: ("OpenType Export Options", EncodedKeyValuesParser, None),
    1744: ("Export Options", ExportOptionsParser, None),
    2001: ("Glyph", GlyphParser, GlyphCompiler),
    2007: ("Background Bitmap", BackgroundBitmapParser, None),
    2008: ("Links", LinkParser, None),
    2009: ("Mask", MaskParser, None),
    2010: ("2010", BaseParser, None),
    2011: ("2011", BaseParser, None),
    2012: ("Mark Color", IntParser, None),
    2013: ("Glyph Bitmaps", GlyphBitmapParser, None),
    2014: ("Binary Tables", BaseParser, None),
    2015: ("Glyph User Data", StringParser, None),
    2016: ("Font User Data", StringParser, None),
    2017: ("Glyph Note", StringParser, None),
    2018: ("Glyph GDEF Data", GlyphGDEFParser, None),
    2020: ("Glyph Anchors Supplemental", GlyphAnchorsSuppParser, None),
    2021: ("Unicode Ranges", IntParser, None),
    2022: ("Export PCLT Table", IntParser, None),
    2023: ("2023", EncodedValueListParser, None),  # Glyph
    2024: ("OpenType Metrics Class Flags", OpenTypeClassFlagsParser, None),
    2025: ("fontNote", StringParser, None),
    2026: ("OpenType Kerning Class Flags", OpenTypeClassFlagsParser, None),
    2027: ("Glyph Origin", GlyphOriginParser, None),
    2028: ("2028", EncodedValueListParser, None),  # MM, proportional to num of masters
    2029: ("Glyph Anchors MM", GlyphAnchorsParser, None),  # MM-compatible
    2031: ("Glyph Guide Properties", GuidePropertiesParser, None),
    2032: ("2032", IntParser, None),
}
# fmt: on

# Make sure the human-readable keys are unique
all_classes = [key for key, _, _ in parser_classes.values()]
assert len(set(all_classes)) == len(all_classes)

entry_ids = {v[0]: k for k, v in parser_classes.items()}

ignore_minimal = [
    "Background Bitmap",
    "fontNote",
    "Global Guides",
    "Glyph Bitmaps",
    "Glyph Guide Properties",
    "Mark Color",
    "Mask",
]



from __future__ import annotations

# try:
#     # NotRequired is only available in Python 3.11+
#     from typing import NotRequired, TypedDict
# except ImportError:
from typing_extensions import NotRequired, TypedDict

from typing import Dict, List, Optional, Tuple
from vfbLib.typing import HintTuple


class TUfoTTZoneDict(TypedDict):
    position: int
    top: bool
    width: int
    # Zone index, shift; index must be str to be saved in lib
    delta: NotRequired[Dict[str, int]]


UfoPoint = Tuple[int, int]
UfoComponent = Tuple[str, Tuple[float, float, float, float, int, int]]
UfoGroups = Dict[str, List[str]]  # name, glyphs
UfoMasterKerning = Dict[Tuple[str, str], int]  # Lstr, Rstr, value
UfoMMKerning = Dict[Tuple[str, int], List[int]]  # Lstr, Rid, master values
UfoSegment = Tuple[Optional[str], bool, Optional[str], UfoPoint]
UfoContour = List[UfoSegment]
TUfoTTZonesDict = Dict[str, TUfoTTZoneDict]


class TUfoGaspRecDict(TypedDict):
    rangeMaxPPEM: int
    rangeGaspBehavior: List[int]


class TUfoStemsDict(TypedDict):
    ttStemsH: List
    ttStemsV: List


class TUfoStemPPMsDict(TypedDict):
    ttStemsH: List
    ttStemsV: List


class UfoGuide(TypedDict):
    angle: float | int
    color: NotRequired[str]
    name: NotRequired[str]
    x: int
    y: int


class UfoHintingV2(TypedDict):
    flexList: NotRequired[List]
    hintSetList: NotRequired[List[UfoHintSet]]
    id: NotRequired[str]


class UfoHintSet(TypedDict):
    pointTag: str
    stems: List[str | HintTuple]



from __future__ import annotations

TT_LIB_KEY = "com.fontlab.v2.tth"
TT_GLYPH_LIB_KEY = "com.fontlab.ttprogram"
PS_GLYPH_LIB_KEY_ADOBE = "com.adobe.type.autohint"
PS_GLYPH_LIB_KEY = "com.adobe.type.autohint.v2"
# TODO: What tool uses the official UFO key?
# PS_GLYPH_LIB_KEY = "public.postscript.hints"


# alignment

ALIGN_CLOSEST_PIXEL = 0
ALIGN_LEFT_BOTTOM = 1
ALIGN_RIGHT_TOP = 2
ALIGN_CENTER = 3
ALIGN_DOUBLE = 4

vfb2ufo_alignment = {
    "default": -1,
    "none": -1,  # Alias
    "round": ALIGN_CLOSEST_PIXEL,
    "bottom": ALIGN_LEFT_BOTTOM,  # Alias
    "left": ALIGN_LEFT_BOTTOM,
    "top": ALIGN_RIGHT_TOP,  # Alias
    "right": ALIGN_RIGHT_TOP,
    "center": ALIGN_CENTER,
    "double": ALIGN_DOUBLE,
}

vfb2ufo_alignment_rev = {
    -1: "default",
    ALIGN_CLOSEST_PIXEL: "round",
    ALIGN_LEFT_BOTTOM: "left",
    ALIGN_RIGHT_TOP: "right",
    ALIGN_CENTER: "center",
    ALIGN_DOUBLE: "double",
}

vfb2ufo_command_codes = {
    "AlignBottom": "alignb",
    "AlignTop": "alignt",
    "SingleLinkH": "singleh",
    "SingleLinkV": "singlev",
    "DoubleLinkH": "doubleh",
    "DoubleLinkV": "doublev",
    "AlignH": "alignh",
    "AlignV": "alignv",
    "InterpolateH": "interpolateh",
    "InterpolateV": "interpolatev",
    "MDeltaH": "mdeltah",
    "MDeltaV": "mdeltav",
    "FDeltaH": "fdeltah",
    "FDeltaV": "fdeltav",
}

vfb2ufo_label_codes = {
    "AlignBottom": "ab",
    "AlignTop": "at",
    "SingleLinkH": "sh",
    "SingleLinkV": "sv",
    "DoubleLinkH": "dh",
    "DoubleLinkV": "dv",
    "AlignH": "ah",
    "AlignV": "av",
    "InterpolateH": "ih",
    "InterpolateV": "iv",
    "MDeltaH": "mh",
    "MDeltaV": "mv",
    "FDeltaH": "fh",
    "FDeltaV": "fv",
    "PSHintReplacement": "hr",
}



from __future__ import annotations

import logging

from fontTools.misc.textTools import hexStr
from functools import cached_property
from io import BytesIO
from struct import pack
from typing import TYPE_CHECKING, Any, Dict, Tuple, Type

# from vfbLib.reader import FALLBACK_PARSER
from vfbLib.compilers import BaseCompiler
from vfbLib.constants import parser_classes
from vfbLib.parsers import BaseParser

if TYPE_CHECKING:
    from io import BufferedReader
    from vfbLib.vfb.vfb import Vfb


logger = logging.getLogger(__name__)


FALLBACK_PARSER = BaseParser


class VfbEntry:
    def __init__(
        self,
        parent: Vfb,
        parser: Type[BaseParser] | None = None,
        compiler: Type[BaseCompiler] | None = None,
    ) -> None:
        # The original or compiled binary data
        self.data: bytes | None = None
        # The decompiled data
        self.decompiled = None
        # Temporary data for additional master, must be merged when compiling
        self.temp_masters = None
        # The numeric and human-readable key of the entry
        self.id = None
        self.key = None
        # Has the data been modified, i.e. it needs recompilation
        self._modified = False
        # The parser which can convert data to decompiled
        self.parser = parser
        # The compiler which can convert the decompiled to compiled data
        self.compiler = compiler
        # The parent object, Vfb
        self.vfb = parent

    @cached_property
    def header(self) -> bytes:
        """The entry header.

        Returns:
            bytes: The data of the current entry header.
        """
        header = BytesIO()
        if self.size > 0xFFFF:
            entry_id = self.id | 0x8000
        else:
            entry_id = self.id
        header.write(pack("<H", entry_id))
        if self.size > 0xFFFF:
            header.write(pack("<I", self.size))
        else:
            header.write(pack("<H", self.size))
        return header.getvalue()

    @cached_property
    def size(self) -> int:
        """The size of the compiled data.

        Returns:
            int: The size of the current compiled data.
        """
        return len(self.data)

    @property
    def modified(self) -> bool:
        return self._modified

    @modified.setter
    def modified(self, value) -> None:
        self._modified = value
        if self._modified:
            try:
                delattr(self, "size")
            except AttributeError:
                pass
        else:
            self.size
        # Optimized version?
        # if value:
        #     if self._modified:
        #         # Value has not changed from True
        #         return

        #     # Value has changed from False to True, invalidate caches
        #     try:
        #         delattr(self, "size")
        #     except AttributeError:
        #         pass
        #     return

        # if self._modified:
        #     # Value changes from True to False
        #     self._modified = False
        #     return

        # # Value is False, no change

    def _read_entry(
        self,
    ) -> Tuple[str, Type[BaseParser], Type[BaseCompiler] | None, int]:
        """
        Read an entry from the stream and return its key, specialized parser
        class, and data size.
        """
        self.id = BaseParser.read_uint16(self.stream)
        entry_info = parser_classes.get(
            self.id & ~0x8000, (str(self.id), FALLBACK_PARSER, None)
        )
        key, parser_class, compiler_class = entry_info

        parser_class.encoding = self.vfb.encoding

        if self.id == 5:
            # File end marker?
            BaseParser.read_uint16(self.stream)
            two = BaseParser.read_uint16(self.stream)
            if two == 2:
                BaseParser.read_uint16(self.stream)
                raise EOFError

        if self.id & 0x8000:
            # Uses uint32 for data length
            num_bytes = BaseParser.read_uint32(self.stream)
        else:
            # Uses uint16 for data length
            num_bytes = BaseParser.read_uint16(self.stream)

        return key, parser_class, compiler_class, num_bytes

    def as_dict(self, minimize=True) -> Dict[str, Any]:
        d: Dict[str, Any] = {
            "key": str(self.key),
        }
        if minimize:
            if self.decompiled is None:
                d["size"] = self.size
                d["data"] = hexStr(self.data)
                if self.parser is not None:
                    d["parser"] = self.parser.__name__
            else:
                d["decompiled"] = self.decompiled
                if self.compiler is not None:
                    d["compiler"] = self.compiler.__name__
                if self.modified:
                    d["modified"] = True
        else:
            d["size"] = self.size
            d["data"] = hexStr(self.data)
            if self.parser is not None:
                d["parser"] = self.parser.__name__
            d["decompiled"] = self.decompiled
            if self.compiler is not None:
                d["compiler"] = self.compiler.__name__
            if self.modified:
                d["modified"] = True
        return d

    def clear_decompiled(self) -> None:
        self.decompiled = None

    def compile(self) -> None:
        """
        Compile the entry. The result is stored in VfbEntry.data.
        """
        if self.compiler is None:
            logger.error(f"Compiling '{self.key}' is not supported yet.")
            return

        self.merge_masters_data()

        self.data = self.compiler().compile(
            self.decompiled, master_count=self.vfb.num_masters
        )
        self.modified = False

    def decompile(self) -> None:
        """
        Decompile the entry. The result is stored in VfbEntry.decompiled.
        """
        if self.decompiled is not None:
            # Already decompiled
            logger.error(
                f"Entry is already decompiled: {self.key}. If you really want to "
                "decompile again, clear the decompiled data beforehand using "
                "`clear_decompiled()`."
            )
            raise ValueError

        if self.parser is None:
            raise ValueError

        if self.data is None:
            raise ValueError

        try:
            self.decompiled = self.parser.parse(
                BytesIO(self.data), size=self.size, master_count=self.vfb.num_masters
            )
        except:  # noqa: E722
            logger.error(f"Parse error for data: {self.key}; {hexStr(self.data)}")
            logger.error(f"Parser class: {self.parser.__name__}")
            self.decompiled = None
            raise

    def merge_masters_data(self) -> None:
        """
        Merge any temporary masters data into the main decompiled structure. Such data
        can be added as the result of drawing with a PointPen into a multiple master
        Vfb.
        """
        if self.temp_masters is None:
            return

        if self.vfb.num_masters == 1:
            return

        if self.compiler is None:
            return

        self.compiler.merge(self.temp_masters, self.decompiled)

    def read(self, stream: BufferedReader) -> None:
        """
        Read the entry from the stream without decompiling the data.
        """
        self.stream = stream
        self.key, self.parser, self.compiler, size = self._read_entry()
        if self.key == "1410":
            # FIXME: Special FL3 stuff?
            if size != 4:
                print(f"Entry 1410 with size {size}")
            self.data = self.stream.read(10)
        else:
            self.data = self.stream.read(size)



from __future__ import annotations

# try:
#     # NotRequired is only available in Python 3.11+
#     from typing import NotRequired, TypedDict
# except ImportError:
from typing_extensions import NotRequired, TypedDict

from typing import Any, Dict, List, Literal, Tuple


Point = Tuple[int, int]


class Anchor(TypedDict):
    name: NotRequired[str]
    x: int
    x1: NotRequired[int]
    y: int
    y1: NotRequired[int]


ClassFlagDict = Dict[str, Tuple[int, int]]


class Component(TypedDict):
    gid: int
    offsetX: List[int]
    offsetY: List[int]
    scaleX: List[float]
    scaleY: List[float]


GaspList = List[Dict[str, int]]


class GdefDict(TypedDict):
    anchors: NotRequired[List[Anchor]]
    carets: NotRequired[List[Tuple[int, int]]]
    glyph_class: NotRequired[str]
    unknown: NotRequired[List[int]]


class GlyphData(TypedDict):
    components: NotRequired[List[Component]]
    # constants: NotRequired[Tuple[Any, ...]]
    guides: NotRequired[GuideDict]
    hints: NotRequired[HintDict]
    imported: NotRequired[Any]  # FIXME
    kerning: NotRequired[Dict[int, List[int]]]
    metrics: NotRequired[List[Point]]
    name: NotRequired[str]
    nodes: NotRequired[List[MMNode]]
    num_masters: NotRequired[int]
    # num_node_values: NotRequired[int]
    tth: NotRequired[List[Instruction]]


class Guide(TypedDict):
    angle: float | int
    pos: int


GuideList = List[Guide]


class GuideDict(TypedDict):
    h: List[GuideList]
    v: List[GuideList]


class GuideProperty(TypedDict):
    color: NotRequired[str]
    index: int
    name: NotRequired[str]


GuidePropertyList = List[GuideProperty]


class Hint(TypedDict):
    pos: int
    width: int


class HintDict(TypedDict):
    h: List[List[Hint]]
    v: List[List[Hint]]
    hintmasks: NotRequired[List[Dict[str, int]]]


HintTuple = Tuple[str, int, int]


class Instruction(TypedDict):
    cmd: str
    params: Dict[str, int]


class LinkDict(TypedDict):
    x: NotRequired[List[Tuple[int, int]]]
    y: NotRequired[List[Tuple[int, int]]]


class MaskData(GlyphData):
    num: int
    reserved0: NotRequired[int]


class MMAnchor(TypedDict):
    x: List[int]
    y: List[int]


class MMHintsDict(TypedDict):
    h: List[List[Hint]]
    v: List[List[Hint]]


class MMNode(TypedDict):
    flags: int
    points: List[List[Point]]
    type: Literal["move", "line", "curve", "qcurve"]


class TTCommandDict(TypedDict):
    name: str
    params: List[str]



from __future__ import annotations

import logging

from vfbLib.parsers import BaseParser
from struct import unpack


logger = logging.getLogger(__name__)


class DoubleParser(BaseParser):
    """
    A parser that reads data as a double-size float.
    """

    @classmethod
    def _parse(cls):
        return unpack("d", cls.stream.read(8))[0]


class FloatListParser(BaseParser):
    """
    A parser that reads data as a list of floats.
    """

    __size__ = 4
    __fmt__ = "f"

    @classmethod
    def _parse(cls):
        values = []
        for _ in range(cls.stream.getbuffer().nbytes // cls.__size__):
            values.extend(unpack(cls.__fmt__, cls.stream.read(cls.__size__)))

        return values


class DoubleListParser(FloatListParser):
    """
    A parser that reads data as a list of doubles.
    """

    __size__ = 8
    __fmt__ = "d"


class IntParser(BaseParser):
    """
    A parser that reads data as UInt16.
    """

    @classmethod
    def _parse(cls):
        return int.from_bytes(cls.stream.read(), byteorder="little", signed=False)


class IntListParser(BaseParser):
    """
    A parser that reads data as a list of UInt16.
    """

    __size__ = 4

    @classmethod
    def _parse(cls):
        values = []
        for _ in range(cls.stream.getbuffer().nbytes // cls.__size__):
            values.append(
                int.from_bytes(
                    cls.stream.read(cls.__size__),
                    byteorder="little",
                    signed=False,
                )
            )
        return values


class PanoseParser(BaseParser):
    """
    A parser that reads data as an array representing PANOSE values.
    """

    @classmethod
    def _parse(cls):
        return unpack("<10b", cls.stream.read())


class SignedIntParser(BaseParser):
    """
    A parser that reads data as signed Int16.
    """

    @classmethod
    def _parse(cls):
        return int.from_bytes(cls.stream.read(), byteorder="little", signed=True)



def binaryToIntList(value: int, start: int = 0):
    intList = []
    counter = start
    while value:
        if value & 1:
            intList.append(counter)
        value >>= 1
        counter += 1
    return intList



from __future__ import annotations

import logging

from io import BytesIO
from typing import Dict, List
from vfbLib.helpers import binaryToIntList
from vfbLib.parsers import BaseParser, read_encoded_value


logger = logging.getLogger(__name__)


class TrueTypeInfoParser(BaseParser):
    """
    A parser that reads data as "TrueType Info" values.
    """

    @classmethod
    def read_key_value_pairs_encoded(
        cls,
        stream: BytesIO,
        num: int,
        target: List,
        key_names: Dict[int, str] | None = None,
    ):
        if key_names is None:
            key_names = {}
        for _ in range(num):
            k = cls.read_uint8(stream)
            v = read_encoded_value(stream)
            target.append({key_names.get(k, str(k)): v})

    @classmethod
    def _parse(cls):
        info_names = {
            0x33: "0x33",
            0x34: "0x34",
            0x35: "0x35",
            0x36: "0x36",
            0x37: "0x37",
            0x38: "0x38",
            0x39: "tt_font_info_settings",  # 0 = false, 65536 = true
            0x3A: "units_per_em",  # duplicate
            0x3B: "0x3b",
            0x3C: "lowest_rec_ppem",
            0x3D: "font_direction_hint",
            0x3E: "weight_class",  # duplicate
            0x3F: "width_class",  # duplicate
            0x40: "embedding",
            0x41: "subscript_x_size",
            0x42: "subscript_y_size",
            0x43: "subscript_x_offset",
            0x44: "subscript_y_offset",
            0x45: "superscript_x_size",
            0x46: "superscript_y_size",
            0x47: "superscript_x_offset",
            0x48: "superscript_y_offset",
            0x49: "strikeout_size",
            0x4A: "strikeout_position",
            0x4B: "ibm_classification",  # ibm_classification + subclass
            0x4C: "OpenTypeOS2Panose",
            0x4D: "OpenTypeOS2TypoAscender",
            0x4E: "OpenTypeOS2TypoDescender",
            0x4F: "OpenTypeOS2TypoLineGap",
            0x50: "0x50",
            0x51: "OpenTypeOS2WinAscent",
            0x52: "OpenTypeOS2WinDescent",
            0x53: "Hdmx PPMs 1",
            0x54: "Codepages",
            0x56: "timestamp",
            0x57: "0x57",
            0x58: "Hdmx PPMs 2",
            0x5C: "Average Width",
        }
        s = cls.stream
        info = []

        while True:
            k = cls.read_uint8(s)

            if k == 0x32:
                return info

            elif k in (0x33, 0x34, 0x35, 0x36, 0x37, 0x38):
                info.append([info_names.get(k, str(k)), read_encoded_value(s)])

            elif k == 0x39:
                bits = binaryToIntList(read_encoded_value(s))
                settings = {
                    16: "use_custom_tt_values",
                    17: "create_vdmx",
                    18: "add_null_cr_space",
                }
                options = [settings.get(i, str(i)) for i in bits]
                info.append([info_names.get(k, str(k)), options])

            elif k in (0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F):
                info.append([info_names.get(k, f"bit{k}"), read_encoded_value(s)])

            elif k in (
                0x40,
                0x41,
                0x42,
                0x43,
                0x44,
                0x45,
                0x46,
                0x47,
                0x48,
                0x49,
                0x4A,
                0x4B,
            ):
                info.append([info_names.get(k, str(k)), read_encoded_value(s)])

            elif k == 0x4C:  # PANOSE?
                v = [cls.read_uint8(s) for _ in range(10)]
                info.append([info_names.get(k, str(k)), v])

            elif k in (0x4D, 0x4E, 0x4F, 0x50, 0x51, 0x52):
                info.append([info_names.get(k, str(k)), read_encoded_value(s)])

            elif k == 0x53:
                num_values = read_encoded_value(s)
                v = [cls.read_uint8(s) for _ in range(num_values)]
                info.append([info_names.get(k, str(k)), v])

            elif k == 0x54:
                # Codepages
                info.append(
                    [
                        info_names.get(k, str(k)),
                        [
                            read_encoded_value(s, signed=False),
                            read_encoded_value(s, signed=False),
                        ],
                    ]
                )

            elif k == 0x56:
                # Timestamp, unsigned
                info.append(
                    [info_names.get(k, str(k)), read_encoded_value(s, signed=False)]
                )

            elif k in (0x57, 0x5C):
                info.append([info_names.get(k, str(k)), read_encoded_value(s)])

            elif k == 0x58:
                num_values = read_encoded_value(s)
                v = [cls.read_uint8(s) for _ in range(num_values)]
                info.append([info_names.get(k, hex(k)), v])

            else:
                logger.info(f"Unknown key in TrueType info: {hex(k)}")


class TrueTypeStemsParser(BaseParser):
    @classmethod
    def _parse(cls):
        stream = cls.stream
        names = ("ttStemsV", "ttStemsH")
        result = {}
        for i in range(2):
            direction = []
            num_stems = read_encoded_value(stream)
            for _ in range(num_stems):
                width = read_encoded_value(stream)
                stem_name_length = cls.read_uint8(stream)
                stem_name = stream.read(stem_name_length).decode("cp1252")
                ppm6 = read_encoded_value(stream)

                direction.append(
                    {
                        "value": width,
                        "name": stem_name,
                        "round": {"6": ppm6},
                    }
                )
            result[names[i]] = direction

        assert stream.read() == b""
        return result


class TrueTypeStemPpemsParser(BaseParser):
    @classmethod
    def _parse(cls):
        stream = cls.stream
        names = ("ttStemsV", "ttStemsH")
        result = {}
        for i in range(2):
            direction = []
            num_stems = read_encoded_value(stream)
            d = {}
            for j in range(num_stems):
                for k in range(2, 6):
                    ppm = read_encoded_value(stream)
                    d[str(k)] = ppm

                direction.append(
                    {
                        "stem": j,
                        "round": d.copy(),
                    }
                )
            result[names[i]] = direction

        assert stream.read() == b""
        return result


class TrueTypeZoneDeltasParser(BaseParser):
    @classmethod
    def _parse(cls):
        stream = cls.stream
        num_deltas = read_encoded_value(stream)
        result = {}
        for _ in range(num_deltas):
            # Index into Bottom + Top Zones
            index = read_encoded_value(stream)
            ppm = read_encoded_value(stream)
            shift = read_encoded_value(stream)
            if index in result:
                result[index][ppm] = shift
            else:
                result[index] = {ppm: shift}

        assert stream.read() == b""
        return result


class TrueTypeZonesParser(BaseParser):
    @classmethod
    def _parse(cls):
        stream = cls.stream
        names = ("ttZonesT", "ttZonesB")
        result = {}
        for i in range(2):
            side = []
            num_zones = read_encoded_value(stream)
            logger.debug(f"Zones: {num_zones}")
            for _ in range(num_zones):
                position = read_encoded_value(stream)
                width = read_encoded_value(stream)
                logger.debug(f"    pos: {position}, width: {width}")
                name_length = read_encoded_value(stream)
                logger.debug(f"Name of length {name_length} follows")
                zone_name = stream.read(name_length).decode("cp1252")
                side.append(
                    {
                        "position": position,
                        "value": width,
                        "name": zone_name,
                    }
                )
            result[names[i]] = side

        assert stream.read() == b""
        return result



from __future__ import annotations

import codecs
import json
import logging

from argparse import ArgumentParser
from pathlib import Path
from vfbLib.ufo import VfbToUfoBuilder
from vfbLib.version import build_date
from vfbLib.vfb.vfb import Vfb


logger = logging.getLogger(__name__)


def vfb2json():
    parser = ArgumentParser(
        description=(
            f"VFB2JSON Converter\nCopyright (c) 2023 by LucasFonts\nBuild {build_date}"
        )
    )
    parser.add_argument(
        "-d",
        "--no-decompile",
        action="store_true",
        default=False,
        help="don't decompile data, output binary in JSON",
    )
    parser.add_argument(
        "--header",
        action="store_true",
        default=False,
        help="only read the VFB header, not the actual data",
    )
    parser.add_argument(
        "-m",
        "--minimal",
        action="store_true",
        default=False,
        help="parse only minimal amount of data",
    )
    parser.add_argument(
        "-p",
        "--path",
        type=str,
        nargs=1,
        help="output folder",
    )
    parser.add_argument(
        "-u",
        "--unicode-strings",
        action="store_true",
        default=False,
        help="interpret name table strings as Unicode instead of Windows-1252",
    )
    parser.add_argument(
        "inputpath",
        type=str,
        nargs=1,
        help="input file path (.vfb)",
    )
    args = parser.parse_args()
    if args:
        vfb_path = Path(args.inputpath[0])
        print(parser.description)
        print(f"Reading file {vfb_path} ...")
        vfb = Vfb(
            vfb_path,
            only_header=args.header,
            minimal=args.minimal,
            unicode_strings=args.unicode_strings,
        )
        if not args.no_decompile:
            vfb.decompile()
        suffix = ".vfb.json"
        if args.path:
            out_path = (Path(args.path[0]) / vfb_path.name).with_suffix(suffix)
        else:
            out_path = vfb_path.with_suffix(suffix)
        with codecs.open(str(out_path), "wb", "utf-8") as f:
            json.dump(vfb.as_dict(), f, ensure_ascii=False, indent=4)
    else:
        parser.print_help()


def vfb2ufo():
    parser = ArgumentParser(
        description=(
            f"VFB3UFO Converter\nCopyright (c) 2023 by LucasFonts\nBuild {build_date}"
        )
    )
    parser.add_argument(
        "-p",
        "--path",
        type=str,
        nargs=1,
        help="output folder",
    )
    parser.add_argument(
        "-fo",
        "--force-overwrite",
        action="store_true",
        default=False,
        help="force overwrite",
    )
    parser.add_argument(
        "-ttx",
        "--ttx",
        action="store_true",
        default=False,
        help="convert binary OpenType Layout data using TTX-like format",
    )
    parser.add_argument(
        "-64",
        "--base64",
        action="store_true",
        default=False,
        help="write GLIF lib 'data' section using base64 (recommended)",
    )
    parser.add_argument(
        "-s",
        "--silent",
        action="store_true",
        default=False,
        help="no display (silent mode)",
    )
    parser.add_argument(
        "-nops",
        "--no-postscript-hints",
        action="store_true",
        default=False,
        help="Don't output PostScript hinting",
    )
    parser.add_argument(
        "-z",
        "--zip",
        action="store_true",
        default=False,
        help="write UFOZ (compressed UFO)",
    )
    parser.add_argument(
        "inputpath",
        type=str,
        nargs=1,
        help="input file path (.vfb)",
    )
    parser.add_argument(
        "outputpath",
        type=str,
        nargs="?",
        help="output file path (.ufo[z])",
    )
    parser.add_argument(
        "-m",
        "--minimal",
        action="store_true",
        default=False,
        help="parse only minimal amount of data, drop missing glyphs from groups, etc.",
    )
    parser.add_argument(
        "-u",
        "--unicode-strings",
        action="store_true",
        default=False,
        help="interpret name table strings as Unicode instead of Windows-1252",
    )
    args = parser.parse_args()
    if args:
        vfb_path = Path(args.inputpath[0])
        if not args.silent:
            print(parser.description)
            print(f"Reading file {vfb_path} ...")
        vfb = Vfb(
            vfb_path,
            minimal=args.minimal,
            drop_keys={"Encoding", "Encoding Mac"},
            unicode_strings=args.unicode_strings,
        )
        suffix = ".ufo"
        if args.zip:
            suffix += "z"
        if args.path:
            out_path = (Path(args.path[0]) / vfb_path.name).with_suffix(suffix)
        else:
            out_path = vfb_path.with_suffix(suffix)
        vfb.decompile()
        builder = VfbToUfoBuilder(
            vfb,
            minimal=args.minimal,
            base64=args.base64,
            pshints=not args.no_postscript_hints,
        )
        builder.write(
            out_path,
            overwrite=args.force_overwrite,
            silent=args.silent,
            ufoz=args.zip,
        )
    else:
        parser.print_help()



from __future__ import annotations

import logging

from functools import cached_property
from typing import TYPE_CHECKING, List, Tuple
from ufoLib2.objects.info import Info, WidthClass
from vfbLib.helpers import binaryToIntList
from vfbLib.ufo.time import convert_timestamp
from vfbLib.ufo.typing import TUfoGaspRecDict


if TYPE_CHECKING:
    from vfbLib.typing import GaspList
    from vfbLib.ufo.typing import UfoGuide


logger = logging.getLogger(__name__)


class VfbToUfoInfo(Info):
    def __init__(self) -> None:
        super(VfbToUfoInfo, self).__init__()
        # Chance to set some defaults that should always be written
        self.familyName = "Untitled"
        self.guidelines: List[UfoGuide] = []
        self.italicAngle: float | int = 0
        self.openTypeGaspRangeRecords: List[TUfoGaspRecDict] = []
        self.openTypeHeadCreated: str | None = None
        self.openTypeHeadFlags: List[int] = []
        self.openTypeHheaLineGap = 0
        self.openTypeHeadLowestRecPPEM = 0
        self.openTypeNameDescription = ""
        self.openTypeNamePreferredFamilyName = ""
        self.openTypeNamePreferredSubfamilyName = ""
        self.openTypeNameRecords: List | None = None
        self.openTypeNameSampleText = ""
        self.openTypeOS2FamilyClass = [0, 0]
        self.openTypeOS2Selection: List[int] = []
        self.openTypeOS2SubscriptXSize = 0
        self.openTypeOS2SubscriptYSize = 0
        self.openTypeOS2SubscriptXOffset = 0
        self.openTypeOS2SubscriptYOffset = 0
        self.openTypeOS2SuperscriptXSize = 0
        self.openTypeOS2SuperscriptYSize = 0
        self.openTypeOS2SuperscriptXOffset = 0
        self.openTypeOS2SuperscriptYOffset = 0
        self.openTypeOS2StrikeoutPosition = 0
        self.openTypeOS2StrikeoutSize = 0
        self.openTypeOS2Type: List[int] = []
        self.openTypeOS2UnicodeRanges: List[int] = []
        self.openTypeOS2WeightClass = 400
        self.openTypeOS2WidthClass = WidthClass(5)
        self.postscriptBlueValues: List[int] = []
        self.postscriptFamilyBlues: List[int] = []
        self.postscriptFamilyOtherBlues: List[int] = []
        self.postscriptFontName = "Untitled-Regular"
        self.postscriptForceBold = False
        self.postscriptFullName = "Untitled Regular"
        self.postscriptIsFixedPitch = False
        self.postscriptOtherBlues: List[int] = []
        self.postscriptStemSnapH: List[int] | None = None
        self.postscriptStemSnapV: List[int] | None = None
        self.postscriptUnderlinePosition = 0
        self.postscriptUnderlineThickness = 0
        self.postscriptWeightName = "Regular"
        self.styleMapFamilyName = "Untitled"
        self.styleMapStyleName = "regular"
        self.styleName = "Regular"

        # Add sensible defaults?
        self.openTypeOS2StrikeoutSize = 0
        self.openTypeOS2StrikeoutPosition = 0
        self.openTypeOS2TypoAscender = 0
        self.openTypeOS2TypoDescender = 0
        self.openTypeOS2TypoLineGap = 0
        self.openTypeOS2WinAscent = 0
        self.openTypeOS2WinDescent = 0
        self.openTypeOS2CodePageRanges: List[int] = []

        self.build_mapping()

    @cached_property
    def ds_family_name(self) -> str | None:
        """
        Return the family name for use in the DesignSpaceDocument.
        """
        if self.openTypeNamePreferredFamilyName:
            return self.openTypeNamePreferredFamilyName
        if self.familyName:
            return self.familyName
        return self.styleMapFamilyName

    @cached_property
    def ds_style_name(self) -> str | None:
        """
        Return the style name for use in the DesignSpaceDocument.
        """
        if self.openTypeNamePreferredSubfamilyName:
            return self.openTypeNamePreferredSubfamilyName
        if self.styleName:
            return self.styleName
        return self.styleMapStyleName

    @cached_property
    def ui_name(self) -> str:
        name = ""
        if hasattr(self, "familyName"):
            name = self.familyName

        elif hasattr(self, "postscriptFontName"):
            name = self.postscriptFontName
        if name:
            return name
        return "Untitled Regular"

    def build_mapping(self):
        self.mapping = {
            # "sgn": "openTypeNamePreferredFamilyName",  # below
            "Menu Name": "styleMapFamilyName",
            "ffn": "postscriptFullName",
            "psn": "postscriptFontName",
            "tfn": "styleMapFamilyName",
            "weight_name": "weightName",
            "Italic Angle": "italicAngle",
            "underlinePosition": "postscriptUnderlinePosition",
            "underlineThickness": "postscriptUnderlineThickness",
            # "Monospaced": "postscriptIsFixedPitch",  # below
            "copyright": "copyright",
            "description": "openTypeNameDescription",
            "manufacturer": "openTypeNameManufacturer",
            "Type 1 Unique ID": "postscriptUniqueID",
            # weight (class), below
            "trademark": "trademark",
            "designer": "openTypeNameDesigner",
            "designerURL": "openTypeNameDesignerURL",
            "manufacturerURL": "openTypeNameManufacturerURL",
            "width_name": "widthName",
            # Default glyph
            "License": "openTypeNameLicense",
            "License URL": "openTypeNameLicenseURL",
            "FOND Family ID": "macintoshFONDFamilyID",
            "FOND Name": "macintoshFONDName",
            "panose": "openTypeOS2Panose",
            "vendorID": "openTypeOS2VendorID",
            "UniqueID": "openTypeNameUniqueID",
            "version": "openTypeNameVersion",
            "versionMajor": "versionMajor",
            "versionMinor": "versionMinor",
            "year": "year",
            "upm": "unitsPerEm",
            # "tsn": "openTypeNamePreferredSubfamilyName",  # below
            "hhea_ascender": "openTypeHheaAscender",
            "hhea_descender": "openTypeHheaDescender",
            "hhea_line_gap": "openTypeHheaLineGap",
            "fontNote": "note",
            "Default Glyph": "postscriptDefaultCharacter",
        }
        # Integer values from TTInfo
        self.mapping_int = {
            # "units_per_em": "unitsPerEm",  # duplicate
            # "weight_class": "openTypeOS2WeightClass",  # duplicate
            "width_class": "openTypeOS2WidthClass",
            "lowest_rec_ppem": "openTypeHeadLowestRecPPEM",
            "subscript_x_size": "openTypeOS2SubscriptXSize",
            "subscript_y_size": "openTypeOS2SubscriptYSize",
            "subscript_x_offset": "openTypeOS2SubscriptXOffset",
            "subscript_y_offset": "openTypeOS2SubscriptYOffset",
            "superscript_x_size": "openTypeOS2SuperscriptXSize",
            "superscript_y_size": "openTypeOS2SuperscriptYSize",
            "superscript_x_offset": "openTypeOS2SuperscriptXOffset",
            "superscript_y_offset": "openTypeOS2SuperscriptYOffset",
            "strikeout_size": "openTypeOS2StrikeoutSize",
            "strikeout_position": "openTypeOS2StrikeoutPosition",
            "OpenTypeOS2TypoAscender": "openTypeOS2TypoAscender",
            "OpenTypeOS2TypoDescender": "openTypeOS2TypoDescender",
            "OpenTypeOS2TypoLineGap": "openTypeOS2TypoLineGap",
            "OpenTypeOS2WinAscent": "openTypeOS2WinAscent",
            "OpenTypeOS2WinDescent": "openTypeOS2WinDescent",
        }

    def fix_underline_position(self):
        # VFB stores middle of line and thickness, but spec says it must be
        # stored as top of line and thickness.
        if (
            self.postscriptUnderlinePosition is None
            or self.postscriptUnderlineThickness is None
        ):
            logger.error(
                "Can't fix underline position because position or thickness is None"
            )
            return

        self.postscriptUnderlinePosition += int(
            round(0.5 * self.postscriptUnderlineThickness)
        )

    def set_attr(self, attr: str, data):
        setattr(self, attr, data)

    def set_created_timestamp(self, data: int) -> None:
        self.openTypeHeadCreated = convert_timestamp(data).strftime("%Y/%m/%d %H:%M:%S")

    def set_fixed_pitch(self, data: int) -> None:
        self.postscriptIsFixedPitch = bool(data)

    def set_name_records(self, data: List[Tuple[int, int, int, int, str]]) -> None:
        self.openTypeNameRecords = []
        for rec in data:
            nameID, platformID, encodingID, languageID, s = rec
            self.openTypeNameRecords.append(
                {
                    "nameID": nameID,
                    "platformID": platformID,
                    "encodingID": encodingID,
                    "languageID": languageID,
                    "string": s,
                }
            )

    def set_selection(self, data: int) -> None:
        # Bit 0 = Italic
        # Bit 5 = Bold
        # Bit 6 = Regular

        # Those should not be included in the list.
        # Any others could remain, but FL doesn't set them.
        intlist = binaryToIntList(data)
        self.openTypeOS2Selection = [b for b in intlist if b not in (0, 5, 6)]

        # Construct the style map style name

        name_parts = []
        if 6 in intlist:
            name_parts.append("regular")
        elif 5 in intlist:
            name_parts.append("bold")
        if 0 in intlist:
            name_parts.append("italic")
        self.styleMapStyleName = " ".join(name_parts)

    def set_style_name(self, data: str) -> None:
        self.styleName = data
        self.openTypeNamePreferredSubfamilyName = data

    def set_tt_gasp(self, data: GaspList) -> None:
        # FIXME: Different results when hinting is compiled with flufocompile
        gasp: List[TUfoGaspRecDict] = []
        for rec in data:
            gasp.append(
                TUfoGaspRecDict(
                    rangeMaxPPEM=rec["maxPpem"],
                    rangeGaspBehavior=binaryToIntList(rec["flags"]),
                )
            )
        self.openTypeGaspRangeRecords = gasp

    def set_tt_info(self, data: List[Tuple[str, int | List[int]]]):
        for k, v in data:
            if isinstance(v, int):
                if k in self.mapping_int:
                    setattr(self, self.mapping_int[k], v)
                elif k == "timestamp":
                    self.set_created_timestamp(v)
                elif k == "font_direction_hint":
                    # self.info.openTypeOS2Type = binaryToIntList(v)
                    pass
                elif k == "embedding":
                    self.openTypeOS2Type = binaryToIntList(v)
                elif k == "ibm_classification":
                    c = v >> 8
                    s = v & ~(c << 8)
                    self.openTypeOS2FamilyClass = [c, s]
                else:
                    logger.info(f"Unhandled integer value in UFO info: {k, v}")
            elif isinstance(v, list):
                if k == "OpenTypeOS2Panose":
                    # Duplicate?
                    # if v != self.info.openTypeOS2Panose:
                    #     print("Contradictory PANOSE values")
                    #     print(self.info.openTypeOS2Panose, "vs.", v)
                    pass
                elif k == "Codepages":
                    cp1, cp2 = v
                    ranges = binaryToIntList(cp1)
                    for cp in binaryToIntList(cp2):
                        ranges.append(cp + 32)
                    if ranges:
                        self.openTypeOS2CodePageRanges = ranges
                else:
                    logger.info(f"Unhandled list value in UFO info: {k, v}")
            else:
                raise TypeError

    def set_unicode_ranges(self, data: int) -> None:
        self.openTypeOS2UnicodeRanges = binaryToIntList(data)

    def set_weight_class(self, data: int) -> None:
        self.openTypeOS2WeightClass = min(max(1, data), 1000)



from __future__ import annotations

import logging

from typing import TYPE_CHECKING, Dict, List, Tuple
from vfbLib.ufo.groups import build_glyph_to_group_maps

if TYPE_CHECKING:
    from vfbLib.ufo.typing import UfoGroups, UfoMasterKerning, UfoMMKerning


logger = logging.getLogger(__name__)


class UfoKerning:
    def __init__(
        self,
        glyph_order: List[str],
        groups: UfoGroups,
        mm_kerning: UfoMMKerning,
    ):
        self.glyph_order = glyph_order
        group_info = build_glyph_to_group_maps(groups)
        self.groups, self.glyph_group_1, self.glyph_group_2 = group_info
        self.mm_kerning = mm_kerning
        self._make_name_based_kerning()

    def _is_exception(self, L: str, R: str):
        L_is_key = f"public.kern1.{L}" in self.groups or L not in self.glyph_group_1

        R_is_key = f"public.kern2.{R}" in self.groups or R not in self.glyph_group_2

        if L_is_key and R_is_key:
            return False

        return True

    def _make_name_based_kerning(self) -> None:
        """
        Convert the glyph indices to glyph names. Also solves group kerning
        references.
        """
        self.mm_kerning_names: Dict[Tuple[str, str], List[int]] = {}
        for pair, values in self.mm_kerning.items():
            L, Rid = pair
            # Make right GID into glyph name
            R = self.glyph_order[int(Rid)]

            # Is the left glyph a keyglyph? It is so if there's a kerning group
            # named after it. In that case, use the group name instead of the
            # glyph name.
            left_group = f"public.kern1.{L}"
            left = left_group if left_group in self.groups else L

            # Is the right glyph a keyglyph?
            right_group = f"public.kern2.{R}"
            right = right_group if right_group in self.groups else R

            self.mm_kerning_names[left, right] = values

    def get_master_kerning(self, master_index: int) -> UfoMasterKerning:
        """
        Extract the kerning values for master_index and return the kerning as
        Dict[Tuple[str, str], int].
        """
        master_kerning: UfoMasterKerning = {}
        for pair, values in self.mm_kerning_names.items():
            value = values[master_index]
            if value != 0 or self._is_exception(*pair):
                master_kerning[pair] = value
        return master_kerning



from __future__ import annotations

import logging

from fontTools.pens.pointPen import (
    AbstractPointPen,
    PointToSegmentPen,
    SegmentToPointPen,
)
from functools import cached_property
from typing import TYPE_CHECKING
from vfbLib.ufo.glyph import VfbToUfoGlyph
from vfbLib.ufo.paths import UfoMasterGlyph
from vfbLib.templates.glyph import get_empty_glyph
from vfbLib.vfb.pens import VfbGlyphPointPen

if TYPE_CHECKING:
    from fontTools.pens.basePen import AbstractPen
    from vfbLib.vfb.vfb import Vfb, VfbMaster
    from vfbLib.vfb.entry import VfbEntry


logger = logging.getLogger(__name__)


class VfbGlyph:
    def __init__(self, entry: VfbEntry, parent: Vfb | VfbMaster) -> None:
        self.entry = entry
        self._parent = parent
        self._glyph: UfoMasterGlyph | None = None
        self.master_index = 0

    # UFO/cu2qu compatibility

    @cached_property
    def name(self) -> str:
        return self._glyph.name

    def clearContours(self):
        try:
            del self.entry.decompiled["hints"]
        except KeyError:
            pass
        self.entry.decompiled["nodes"] = []
        self.entry.modified = True

    # Native methods

    def decompile(self) -> str:
        """
        Decompile the Glyph entry and return the glyph name.
        """
        self.entry.decompile()
        if self.entry.decompiled is None:
            raise ValueError

        return self.entry.decompiled["name"]

    def empty(self):
        self.entry.decompiled = get_empty_glyph(self._parent.num_masters)

    def _copy_to_ufo_glyph(self):
        """
        Copy minimal data to the VfbToUfoGlyph. Only data that is necessary for the pen
        methods is copied.
        """
        if self.entry.decompiled is None:
            raise ValueError

        _mm_glyph = VfbToUfoGlyph()
        _mm_glyph.name = self.entry.decompiled["name"]

        if "components" in self.entry.decompiled:
            _mm_glyph.mm_components = self.entry.decompiled["components"]

        if "metrics" in self.entry.decompiled:
            _mm_glyph.mm_metrics = self.entry.decompiled["metrics"]

        if "nodes" in self.entry.decompiled:
            _mm_glyph.mm_nodes = self.entry.decompiled["nodes"]

        # TODO: Support point names used by TT hinting
        # if "tth" in self.entry.decompiled:
        #     _mm_glyph.tth_commands = self.entry.decompiled["tth"]

        self._glyph = UfoMasterGlyph(
            mm_glyph=_mm_glyph,
            glyph_order=self._parent.glyph_order,
            master_index=self._parent.master_index,
        )
        self._glyph.build()

    def draw(self, pen) -> None:
        """
        Draw the VFB glyph onto a segment pen.
        """
        sp = PointToSegmentPen(pen, outputImpliedClosingLine=False)
        self.drawPoints(sp)

    def drawPoints(self, pen: AbstractPointPen) -> None:
        """
        Draw the VFB glyph onto a point pen.
        """
        if self.entry.decompiled is None:
            raise ValueError

        if self._glyph is None:
            self._copy_to_ufo_glyph()

        if self._glyph is None:
            raise ValueError

        return self._glyph.drawPoints(pen)

    def getPen(self) -> AbstractPen:
        """
        Return a segment pen to draw into the VFB glyph.
        """
        # TODO: Test
        return SegmentToPointPen(self.getPointPen(), guessSmooth=True)

    def getPointPen(self) -> VfbGlyphPointPen:
        """
        Return a point pen to draw into the VFB glyph.
        """
        if self.entry.decompiled is None:
            if self.entry.data is None:
                # Make an empty glyph
                self.empty(self._parent.num_masters)
            else:
                self.decompile()
        return VfbGlyphPointPen(self, self._parent)


class VfbGlyphMaster:
    def __init__(self, glyph: VfbGlyph, master_index: int = 0):
        self.glyph = glyph
        self.entry = glyph.entry
        self.master_index = master_index
        self._glyph = None

    @cached_property
    def name(self) -> str:
        return self._glyph.name

    def _copy_to_ufo_glyph(self, master_index):
        """
        Copy minimal data to the VfbToUfoGlyph. Only data that is necessary for the pen
        methods is copied.
        """
        if self.entry.decompiled is None:
            raise ValueError

        _mm_glyph = VfbToUfoGlyph()
        _mm_glyph.name = self.entry.decompiled["name"]

        if "components" in self.entry.decompiled:
            _mm_glyph.mm_components = self.entry.decompiled["components"]

        if "metrics" in self.entry.decompiled:
            _mm_glyph.mm_metrics = self.entry.decompiled["metrics"]

        if "nodes" in self.entry.decompiled:
            _mm_glyph.mm_nodes = self.entry.decompiled["nodes"]

        # TODO: Support point names used by TT hinting
        # if "tth" in self.entry.decompiled:
        #     _mm_glyph.tth_commands = self.entry.decompiled["tth"]

        self._glyph = UfoMasterGlyph(
            mm_glyph=_mm_glyph,
            glyph_order=self.glyph._parent.glyph_order,
            master_index=master_index,
        )
        self._glyph.build()

    def clearContours(self):
        if self.master_index == 0:
            target = self.entry.decompiled
        else:
            if self.entry.temp_masters is None:
                self.entry.temp_masters = [
                    {} for _ in range(self.glyph._parent.num_masters)
                ]
            target = self.entry.temp_masters[self.master_index]
        try:
            del target["hints"]
        except KeyError:
            pass
        target["nodes"] = []
        self.entry.modified = True

    def drawPoints(self, pen: AbstractPointPen) -> None:
        """
        Draw the VFB glyph onto a point pen.
        """
        if self.entry.decompiled is None:
            raise ValueError

        if self._glyph is None:
            self._copy_to_ufo_glyph(self.master_index)

        if self._glyph is None:
            raise ValueError

        return self._glyph.drawPoints(pen)

    def getPen(self) -> AbstractPen:
        """
        Return a segment pen to draw into the VFB glyph.
        """
        # TODO: Test
        return SegmentToPointPen(self.getPointPen(), guessSmooth=True)

    def getPointPen(self) -> VfbGlyphPointPen:
        """
        Return a point pen to draw into the VFB glyph.
        """
        if self.entry.decompiled is None:
            if self.entry.data is None:
                # Make an empty glyph
                self.empty(self.glyph._parent.num_masters)
            else:
                self.decompile()
        return VfbGlyphPointPen(self, self.glyph._parent)



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List

if TYPE_CHECKING:
    from fontTools.designspaceLib import (
        AxisDescriptor,
        DiscreteAxisDescriptor,
    )


def get_ds_location(
    axes: List[AxisDescriptor | DiscreteAxisDescriptor],
    vfb_location: List[float],
    factor=1000,
) -> Dict[str, float]:
    # Transfrom a VFB master or instance location to a designspace location dict.
    # Master locations are given in VFB as per-mille in the design coords, so we need
    # the factor 1000.
    # For instance locations, the factor must be 1.
    ds_location: Dict[str, float] = {}
    for i in range(len(axes)):
        axis = axes[i]
        if axis.name:
            ds_location[axis.name] = factor * vfb_location[i]
        else:
            raise ValueError(f"Axis doesn't have a name: {axis}")
    return ds_location


def get_ds_design_location(
    axes: List[AxisDescriptor | DiscreteAxisDescriptor], vfb_location: List[float]
) -> Dict[str, float]:
    # Transform an instance user location to a design location.
    # Instances are given in the VFB as user coords, but as design coords in the DS.
    ds_user_location = get_ds_location(axes, vfb_location, 1)
    axis_names = {axis.name: axis for axis in axes}
    ds_design_location = {
        name: axis_names[name].map_forward(loc)
        for name, loc in ds_user_location.items()
    }
    return ds_design_location



from __future__ import annotations

import codecs
import logging

from defcon.objects.font import Font
from ufonormalizer import normalizeUFO
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import List, Literal
from vfbLib.ufo.pshints import update_adobe_hinting
from vfbLib.ufo.vfb2ufo import (
    PS_GLYPH_LIB_KEY_ADOBE,
    PS_GLYPH_LIB_KEY,
    TT_GLYPH_LIB_KEY,
)


logger = logging.getLogger(__name__)

RF_GUIDES_KEY = "com.typemytype.robofont.guides"

delete_lib_keys: List[str] = []


def fix_vfb2ufo_feature_encoding(ufo_path: Path) -> None:
    # Read vfb2ufo's Windows-1252-encoded feature file and convert to UTF-8
    fea_path = ufo_path / "features.fea"
    if not fea_path.exists():
        # UFOs may omit the fea file
        return

    with codecs.open(str(fea_path), "rb", "cp1252") as f:
        fea = f.read()
    with codecs.open(str(fea_path), "wb", "utf-8") as f:
        f.write(fea)


def update_global_guides(f: Font):
    # Update Global Guides to UFO standard
    if RF_GUIDES_KEY in f.lib:
        guides = []
        for guide in f.lib[RF_GUIDES_KEY]:
            isGlobal = guide["isGlobal"]
            del guide["isGlobal"]
            del guide["magnetic"]
            assert isGlobal
            guides.append(guide)
        if guides:
            f.guidelines = guides
        del f.lib[RF_GUIDES_KEY]


def update_glyph_guides(glyph):
    # Update Guides to UFO standard
    if RF_GUIDES_KEY in glyph.lib:
        guides = []
        for guide in glyph.lib[RF_GUIDES_KEY]:
            isGlobal = guide["isGlobal"]
            del guide["isGlobal"]
            del guide["magnetic"]
            assert not isGlobal
            guides.append(guide)
        if guides:
            glyph.guidelines = guides
        del glyph.lib[RF_GUIDES_KEY]


def normalize_ufo(
    filepath: Path, structure: Literal["package", "zip"] = "package"
) -> None:
    logger.info(f"Processing {filepath.name}...")

    normalized_file = filepath / ".normalized"

    if structure == "package":

        if normalized_file.exists():
            logger.info(f"    Skipping already normalized UFO: {filepath.name}")
            return

        fix_vfb2ufo_feature_encoding(filepath)  # FIXME: Support ufoz

    try:
        f = Font(str(filepath))
    except:  # noqa: E722
        logger.error(f"Skipping UFO with errors: {filepath}")
        raise

    with NamedTemporaryFile(suffix="ufoz") as tf:
        f.save(path=tf.name, formatVersion=3, structure="zip")
        f = Font(tf.name)
        for glyph in f:
            for key in delete_lib_keys:
                try:
                    del glyph.lib[key]
                except KeyError:
                    pass

            # Make TT glyph program readable
            if TT_GLYPH_LIB_KEY in glyph.lib:
                data = glyph.lib[TT_GLYPH_LIB_KEY]
                try:
                    data = data.decode()
                except AttributeError:
                    pass
                glyph.lib[TT_GLYPH_LIB_KEY] = data

            # Update PS Hinting to V2
            if PS_GLYPH_LIB_KEY_ADOBE in glyph.lib:
                v2 = update_adobe_hinting(glyph.lib[PS_GLYPH_LIB_KEY_ADOBE])
                if v2:
                    glyph.lib[PS_GLYPH_LIB_KEY] = v2
                del glyph.lib[PS_GLYPH_LIB_KEY_ADOBE]

            update_glyph_guides(glyph)

        update_global_guides(f)

        if structure == "zip":
            if not filepath.suffix == ".ufoz":
                filepath = filepath.with_suffix(".ufoz")
        else:
            if filepath.suffix == ".ufoz":
                filepath = filepath.with_suffix(".ufo")
        f.save(path=str(filepath), formatVersion=3, structure=structure)
        normalizeUFO(ufoPath=str(filepath), onlyModified=False, writeModTimes=False)
        if structure == "package":
            normalized_file.touch()


def normalize_ufoz(filepath):
    """
    Normalize the ufo, but save as .ufoz
    """
    normalize_ufo(filepath, structure="zip")


def normalize():
    from sys import argv

    for ufo_path in argv[1:]:
        normalize_ufo(Path(ufo_path))



from __future__ import annotations

import logging

from fontTools.misc.textTools import hexStr
from io import BufferedReader, BytesIO
from struct import unpack
from typing import TYPE_CHECKING, Any, Dict, List
from vfbLib.parsers.value import read_doubles, read_encoded_value, read_floats

if TYPE_CHECKING:
    from vfbLib.typing import ClassFlagDict


logger = logging.getLogger(__name__)

uint8 = 1
uint16 = 2
uint32 = 4


class BaseParser:
    """
    Base class to parse data from a vfb file
    """

    encoding = "cp1252"
    master_count: int | None = None
    stream: BytesIO = BytesIO()

    @classmethod
    def parse(
        cls,
        stream: BufferedReader,
        size: int,
        master_count: int | None = None,
    ):
        cls.stream = BytesIO(stream.read(size))
        cls.master_count = master_count
        return cls._parse()

    @classmethod
    def _parse(cls) -> Any:
        return hexStr(cls.stream.read())

    @classmethod
    def read_double(cls, stream=None):
        if stream is None:
            stream = cls.stream
        return read_doubles(1, stream)[0]

    @classmethod
    def read_doubles(cls, num, stream=None):
        if stream is None:
            stream = cls.stream
        return read_doubles(num, stream)

    @classmethod
    def read_float(cls, stream=None):
        if stream is None:
            stream = cls.stream
        return read_floats(1, stream)[0]

    @classmethod
    def read_floats(cls, num, stream=None):
        if stream is None:
            stream = cls.stream
        return read_floats(num, stream)

    @classmethod
    def read_int16(cls, stream=None) -> int:
        if stream is None:
            stream = cls.stream
        return int.from_bytes(stream.read(uint16), byteorder="little", signed=True)

    @classmethod
    def read_int32(cls, stream=None) -> int:
        if stream is None:
            stream = cls.stream
        return int.from_bytes(stream.read(uint32), byteorder="little", signed=True)

    @classmethod
    def read_uint8(cls, stream=None) -> int:
        if stream is None:
            stream = cls.stream
        return int.from_bytes(stream.read(uint8), byteorder="little", signed=False)

    @classmethod
    def read_uint16(cls, stream=None) -> int:
        if stream is None:
            stream = cls.stream
        return int.from_bytes(stream.read(uint16), byteorder="little", signed=False)

    @classmethod
    def read_uint32(cls, stream=None) -> int:
        if stream is None:
            stream = cls.stream
        return int.from_bytes(stream.read(uint32), byteorder="little", signed=False)


class EncodedKeyValuesParser(BaseParser):
    __end__ = 0x64

    @classmethod
    def _parse(cls) -> List[Dict[int, int]]:
        values = []
        while True:
            key = cls.read_uint8()
            if key == cls.__end__:
                break

            val = read_encoded_value(cls.stream)
            values.append({key: val})

        return values


class EncodedKeyValuesParser1742(EncodedKeyValuesParser):
    __end__ = 0x00


class EncodedValueParser(BaseParser):
    """
    A parser that reads data as Yuri's optimized encoded value (1 value).
    """

    @classmethod
    def _parse(cls) -> int:
        value = read_encoded_value(cls.stream)
        assert cls.stream.read() == b""
        return value


class EncodedValueListParser(BaseParser):
    """
    A parser that reads data as Yuri's optimized encoded values.
    """

    @classmethod
    def _parse(cls) -> List[int]:
        values = []
        while True:
            try:
                val = read_encoded_value(cls.stream)
                values.append(val)
            except EOFError:
                logger.debug("EOF")
                return values


class EncodedValueListWithCountParser(BaseParser):
    """
    A parser that reads data as Yuri's optimized encoded values. The list of values is
    preceded by a count value that specifies how many values should be read.
    """

    @classmethod
    def _parse(cls) -> Dict[str, List[int]]:
        count = read_encoded_value(cls.stream)
        values: Dict[str, List[int]] = {"values": []}
        for _ in range(count):
            val = read_encoded_value(cls.stream)
            values["values"].append(val)
        return values


class GaspParser(BaseParser):
    """
    A parser that reads data as an array representing Gasp table values.
    """

    @classmethod
    def _parse(cls):
        data = cls.stream.read()
        # cls.stream.getbuffer().nbytes
        gasp = unpack(f"<{len(data) // 2}H", data)
        it = iter(gasp)
        return [
            {
                "maxPpem": a,
                "flags": b,
            }
            for a, b in zip(it, it)
        ]


class GlyphEncodingParser(BaseParser):
    @classmethod
    def _parse(cls):
        gid = int.from_bytes(cls.stream.read(2), byteorder="little")
        nam = cls.stream.read().decode("cp1252")
        return gid, nam


class OpenTypeClassFlagsParser(BaseParser):
    @classmethod
    def _parse(cls) -> ClassFlagDict:
        class_flags: ClassFlagDict = {}
        num_classes = read_encoded_value(cls.stream)
        for _ in range(num_classes):
            n = read_encoded_value(cls.stream)
            name = cls.stream.read(n).decode(cls.encoding)
            flag1 = read_encoded_value(cls.stream)
            flag2 = read_encoded_value(cls.stream)
            class_flags[name] = (flag1, flag2)
        return class_flags



from __future__ import annotations

from typing import Dict
from vfbLib.typing import TTCommandDict


TT_COMMANDS: Dict[int, TTCommandDict] = {
    0x01: {"name": "AlignTop", "params": ["pt", "zone"]},
    0x02: {"name": "AlignBottom", "params": ["pt", "zone"]},
    0x03: {
        "name": "SingleLinkH",
        "params": ["pt1", "pt2", "stem", "align"],
    },
    0x04: {
        "name": "SingleLinkV",
        "params": ["pt1", "pt2", "stem", "align"],
    },
    0x05: {
        "name": "DoubleLinkH",
        "params": ["pt1", "pt2", "stem"],
    },
    0x06: {
        "name": "DoubleLinkV",
        "params": ["pt1", "pt2", "stem"],
    },
    0x07: {"name": "AlignH", "params": ["pt", "align"]},
    0x08: {"name": "AlignV", "params": ["pt", "align"]},
    0x0D: {
        "name": "InterpolateH",
        "params": ["pti", "pt1", "pt2", "align"],
    },
    0x0E: {
        "name": "InterpolateV",
        "params": ["pti", "pt1", "pt2", "align"],
    },
    0x14: {
        "name": "MDeltaH",
        "params": ["pt", "shift", "ppm1", "ppm2"],
    },
    0x15: {
        "name": "MDeltaV",
        "params": ["pt", "shift", "ppm1", "ppm2"],
    },
    0x16: {
        "name": "FDeltaH",
        "params": ["pt", "shift", "ppm1", "ppm2"],
    },
    0x17: {
        "name": "FDeltaV",
        "params": ["pt", "shift", "ppm1", "ppm2"],
    },
}

# TT command names to constants
TT_COMMAND_CONSTANTS = {v["name"]: k for k, v in TT_COMMANDS.items()}



from __future__ import annotations

import logging

from typing import TYPE_CHECKING, Any, Dict, List
from vfbLib.ufo.glyph import VfbToUfoGlyph
from vfbLib.ufo.vfb2ufo import (
    TT_GLYPH_LIB_KEY,
    vfb2ufo_alignment_rev,
    vfb2ufo_command_codes,
)

if TYPE_CHECKING:
    from vfbLib.ufo.typing import TUfoStemsDict


logger = logging.getLogger(__name__)


def get_xml_tth(commands) -> List[str]:
    """
    Convert the glyph's list of TTH commands to a list of TTH command xml strings.
    """
    return [tt_cmd_dict_to_xml(cmd_dict) for cmd_dict in commands]


def set_tth_lib(glyph, commands) -> None:
    """
    Save the TTH commands to the glyph's lib. Optionally rename the hinted points.
    """
    tth = get_xml_tth(commands)
    if tth:
        glyph.lib[TT_GLYPH_LIB_KEY] = (
            "  <ttProgram>\n" + "\n".join(tth) + "\n  </ttProgram>\n"
        )


def tt_cmd_dict_to_xml(tt_dict: Dict[str, Any]) -> str:
    """
    Convert the dict tt command into a FontLab XML string.
    """
    code = tt_dict["code"]
    cmd = f'    <ttc code="{code}"'
    for attr in (
        "point",
        "point1",
        "point2",
        "round",
        "stem",
        "zone",
        "align",
        "delta",
        "ppm1",
        "ppm2",
    ):
        if attr in tt_dict:
            if attr == "round":
                val = str(tt_dict[attr]).lower()
            else:
                val = tt_dict[attr]
            cmd += f' {attr}="{val}"'
    cmd += "/>"
    return cmd


def transform_stem_rounds(data: Dict[str, int], name: str) -> Dict[str, int]:
    d = {"0": 1}
    for k, v in data.items():
        key = str(v)
        val = int(k)
        if key in d:
            logger.error(
                f"Error in stem rounding settings for {name}, duplicate ppm {key}."
            )
        d[key] = val
    return d


class TTGlyphHints:
    def __init__(
        self,
        mm_glyph: VfbToUfoGlyph,
        data: List[Dict[str, Any]],
        zone_names: Dict[str, List[str]],
        stems: TUfoStemsDict,
    ) -> None:
        self.glyph: VfbToUfoGlyph = mm_glyph
        self.zone_names = zone_names
        self.stems = stems

        self.commands: List[Dict] = []

        self._build_tt_glyph_hints(data)

    def _build_tt_glyph_hints(self, data: List[Dict[str, Any]]) -> None:
        # Write TT hints into glyph lib.
        self.commands = []
        for cmd in data:
            code = cmd["cmd"]
            params = cmd["params"]
            d: Dict[str, str | bool] = {"code": vfb2ufo_command_codes[code]}
            if code in ("AlignBottom", "AlignTop"):
                d["point"] = self.glyph.get_point_label(params["pt"], code)
                if code == "AlignBottom":
                    zd = "ttZonesB"
                else:
                    zd = "ttZonesT"
                d["zone"] = self.zone_names[zd][params["zone"]]
            elif code in ("AlignH", "AlignV"):
                d["point"] = self.glyph.get_point_label(params["pt"], code)
                if "align" in params:
                    align = params["align"]
                    if align > -1:
                        d["align"] = vfb2ufo_alignment_rev[align]
            elif code in (
                "SingleLinkH",
                "SingleLinkV",
                "DoubleLinkH",
                "DoubleLinkV",
            ):
                d["point1"] = self.glyph.get_point_label(params["pt1"], code)
                d["point2"] = self.glyph.get_point_label(params["pt2"], code)
                if "stem" in params:
                    stem = params["stem"]
                    if stem <= -2:
                        d["round"] = True
                    elif stem == -1:
                        pass
                    else:
                        stem_dir = "ttStemsH" if code.endswith("H") else "ttStemsV"
                        if stem >= len(self.stems[stem_dir]):
                            logger.warning(
                                f"Stem index in {stem_dir} out of range in "
                                f"{self.glyph.name}: {stem} (of "
                                f"{len(self.stems[stem_dir])} existing stems). "
                                "Choosing first stem."
                            )
                            logger.warning(f"{code}: {params}")
                            logger.warning(self.stems[stem_dir])
                            stem = 0
                        d["stem"] = self.stems[stem_dir][stem]["name"]
                if "align" in params:
                    align = params["align"]
                    if align > -1:
                        d["align"] = vfb2ufo_alignment_rev[align]
            elif code in (
                "InterpolateH",
                "InterpolateV",
            ):
                d["point"] = self.glyph.get_point_label(params["pti"], code)
                d["point1"] = self.glyph.get_point_label(params["pt1"], code)
                d["point2"] = self.glyph.get_point_label(params["pt2"], code)
                if "align" in params:
                    align = params["align"]
                    if align > -1:
                        d["align"] = vfb2ufo_alignment_rev[align]
            elif code in (
                "MDeltaH",
                "MDeltaV",
                "FDeltaH",
                "FDeltaV",
            ):
                d["point"] = self.glyph.get_point_label(params["pt"], code)
                d["delta"] = params["shift"]
                d["ppm1"] = params["ppm1"]
                d["ppm2"] = params["ppm2"]
            else:
                logger.error(f"Unknown TT command: {code}")
                raise ValueError

            self.commands.append(d)



from __future__ import annotations

import logging

from pathlib import Path
from time import time
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Set, Tuple
from vfbLib.vfb.glyph import VfbGlyph, VfbGlyphMaster
from vfbLib.vfb.entry import VfbEntry
from vfbLib.vfb.header import VfbHeader
from vfbLib.vfb.info import VfbInfo

if TYPE_CHECKING:
    from io import BufferedReader


logger = logging.getLogger(__name__)


# Convenience objects for vfb access


class Vfb:
    """
    Object to represent the vfb data, with the ability to read and write. You can use
    the Vfb object to access glyphs through dict methods, where the glyph name is the
    key and the glyph object is the value.
    """

    def __init__(
        self,
        vfb_path: Path,
        timing=True,
        minimal=False,
        drop_keys: Set[str] | None = None,
        only_header=False,
        unicode_strings = False,
    ) -> None:
        self.vfb_path = vfb_path
        self.timing = timing
        self.minimal = minimal
        if drop_keys is None:
            self.drop_keys = set()
        else:
            self.drop_keys = set(drop_keys)
        self.only_header = only_header
        # String encoding for nametable entries
        self.encoding = "utf-8" if unicode_strings else "cp1252"

        # We need some minimal API to make pen access work ...
        self._glyphs: Dict[str, VfbGlyph] = {}
        self.glyph_order: List[str] = []

        # cu2qu accesses the info and lib ...
        self.info = VfbInfo(vfb=self)
        self.lib = {}

        self.master_index = 0
        self.num_masters: int = 0
        self.read()

    def as_dict(self) -> Dict[str, Dict[str, Any]]:
        """
        Return the Vfb structure as Dict, e.g. for saving as JSON. The dict has the keys
        "header" and "entries".
        """
        d = {}
        if self.header is not None:
            d["header"] = self.header.as_dict()
        if self.entries:
            d["entries"] = [e.as_dict() for e in self.entries]
        return d

    def clear(self):
        """
        Clear any data that may have been read before.
        """
        self.header: VfbHeader | None = None
        self.entries: List[VfbEntry] = []

    def _decompile_glyphs(self):
        for entry in self.entries:
            if entry.key == "Glyph":
                glyph = VfbGlyph(entry, self)
                name = glyph.decompile()
                i = 0
                while name in self._glyphs:
                    logger.error(f"VFB contains duplicate glyph name: {name}")
                    # Disambiguate duplicate names
                    i += 1
                    name = f"{name}#{i}"
                self._glyphs[name] = glyph
                self.glyph_order.append(name)

    def __contains__(self, key: str) -> bool:
        if not self._glyphs:
            self._decompile_glyphs()
        return key in self._glyphs

    def __getitem__(self, key: str) -> VfbGlyph:
        if not self._glyphs:
            self._decompile_glyphs()
        return self._glyphs[key]

    def getGlyphMaster(self, key: str, master_index: int) -> VfbGlyphMaster:
        return VfbGlyphMaster(self[key], master_index)

    def get_masters(self) -> List[VfbMaster]:
        return [VfbMaster(self, i) for i in range(self.num_masters)]

    def decompile(self) -> None:
        """
        Decompile all entries, except for the ones listed in `drop_keys`.
        """
        start = time()
        for entry in self.entries:
            entry.decompile()

        end = time()
        if self.timing:
            print(f"Interpreting binary data took {round((end - start) * 1000)} ms.")

    def items(self) -> Iterable[Tuple[str, VfbGlyph]]:
        if not self._glyphs:
            self._decompile_glyphs()
        return self._glyphs.items()

    def keys(self) -> Iterable[str]:
        if not self._glyphs:
            self._decompile_glyphs()
        return self._glyphs.keys()

    def read_stream(self, stream: BufferedReader):
        """
        Lazily read and parse the vfb stream, i.e. parse the header, but only read the
        binary data of other entries.
        """
        start = time()
        self.header = VfbHeader()
        self.header.read(stream)
        if self.only_header:
            return

        entry: VfbEntry | None = None
        while True:
            try:
                entry = VfbEntry(self)
                entry.read(stream)
            except EOFError:
                break

            if entry is not None:
                if entry.key == "Master Count":
                    entry.decompile()
                    self.num_masters = entry.decompiled
                    entry.decompiled = None

                if entry.key not in self.drop_keys:
                    self.entries.append(entry)

        end = time()
        if self.timing:
            print(
                "Source file was successfully read in "
                f"{round((end - start) * 1000)} ms."
            )

    def read(self):
        """
        Read data from the file at vfb_path, without decompiling
        """
        self.clear()
        with open(self.vfb_path, "rb") as vfb:
            self.read_stream(vfb)

    def write(self, out_path: Path) -> None:
        """
        Compile any entries with changes, and write the VFB to out_path.
        """
        if self.header is None:
            raise ValueError

        with open(out_path, "wb") as vfb:
            if self.header.modified:
                self.header.compile()
            assert self.header.data
            vfb.write(self.header.data)

            for entry in self.entries:
                if entry.modified:
                    entry.compile()
                vfb.write(entry.header)
                vfb.write(entry.data)
            # File end marker
            vfb.write(b"\05\00\00\00\02\00\00\00")


class VfbMaster:
    """
    Minimal UFO interface for a single master of a multiple master VFB.
    """

    def __init__(self, vfb: Vfb, master_index: int = 0):
        self.vfb = vfb
        self.master_index = master_index

    def __contains__(self, key: str) -> bool:
        return key in self.vfb

    def __getitem__(self, key: str) -> VfbGlyphMaster:
        return self.vfb.getGlyphMaster(key, self.master_index)

    @property
    def info(self) -> VfbInfo:
        return self.vfb.info

    def items(self) -> Iterable[Tuple[str, VfbGlyph]]:
        return self.vfb.items()

    def keys(self) -> Iterable[str]:
        return self.vfb.keys()

    def num_masters(self) -> int:
        return self.vfb.num_masters



from __future__ import annotations

from argparse import ArgumentParser
from fontTools.cu2qu.ufo import font_to_quadratic, fonts_to_quadratic
from pathlib import Path
from vfbLib.version import build_date
from vfbLib.vfb.vfb import Vfb


def vfbcu2qu():
    parser = ArgumentParser(
        description=(
            "VFB Cubic to Quadratic Converter\nCopyright (c) 2023 by LucasFonts\n"
            f"Build {build_date}"
        )
    )
    parser.add_argument(
        "-p",
        "--path",
        type=str,
        nargs=1,
        help="output folder",
    )
    parser.add_argument(
        "-fo",
        "--force-overwrite",
        action="store_true",
        default=False,
        help="force overwrite",
    )
    parser.add_argument(
        "inputpath",
        type=str,
        nargs=1,
        help="input file path (.vfb)",
    )
    parser.add_argument(
        "outputpath",
        type=str,
        nargs="?",
        help="output file path (.vfb)",
    )
    parser.add_argument(
        "-m",
        "--max-err-em",
        type=float,
        nargs=1,
        help="Maximum allowed error, relative to the font's units per em.",
    )
    args = parser.parse_args()
    if args:
        vfb_path = Path(args.inputpath[0])
        print(parser.description)
        print(f"Reading file {vfb_path} ...")
        vfb = Vfb(vfb_path, drop_keys={"Links"})
        kwargs = {
            "max_err_em": None,
            "max_err": None,
            "reverse_direction": True,
            "stats": None,
            "dump_stats": False,
            "remember_curve_type": False,  # Prevent write access to the lib
            "all_quadratic": True,
        }
        if args.max_err_em:
            kwargs["max_err_em"] = args.max_err_em[0]
        if vfb.num_masters == 1:
            modified = font_to_quadratic(vfb, **kwargs)
        elif vfb.num_masters > 1:
            vfbs = vfb.get_masters()
            modified = fonts_to_quadratic(vfbs, **kwargs)
        else:
            print(f"Unsupported number of masters: {vfb.num_masters}")
            return

        if not modified:
            print("File was not modified.")
            return

        suffix = ".qu.vfb"
        if args.path:
            suffix = ".vfb"
            out_path = (Path(args.path[0]) / vfb_path.name).with_suffix(suffix)
        else:
            out_path = vfb_path.with_suffix(suffix)
        if out_path.exists():
            if not args.force_overwrite:
                print(
                    "Output file exists, new file was not saved. "
                    "Use -fo to force overwriting."
                )
                raise FileExistsError

        print(f"Saving converted file to {out_path}.")
        vfb.write(out_path)
    else:
        parser.print_help()



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from vfbLib.vfb.vfb import Vfb


class VfbInfo:
    def __init__(self, vfb: Vfb) -> None:
        self.parent = vfb

    @cached_property
    def unitsPerEm(self) -> int:
        for entry in self.parent.entries:
            if entry.key == "upm":
                entry.decompile()
                if entry.decompiled is None:
                    return 0
                return entry.decompiled
        return 0



from __future__ import annotations

import logging

from typing import TYPE_CHECKING, List
from vfbLib.ufo.typing import UfoGuide

if TYPE_CHECKING:
    from vfbLib.typing import (
        GuideDict,
        GuidePropertyList,
    )


logger = logging.getLogger(__name__)


def get_master_guides(mm_guides: GuideDict, master_index: int) -> List[UfoGuide]:
    # Concatenate guidlines for both directions and extract coords for
    # master_index
    guides = []
    for d in ("h", "v"):
        direction_mm_guides = mm_guides[d]
        for master_guide in direction_mm_guides[master_index]:
            guide = UfoGuide(angle=0, x=0, y=0)
            if d == "h":
                guide["y"] = master_guide["pos"]
            else:
                guide["x"] = master_guide["pos"]
            angle = master_guide["angle"]
            if d == "v":
                angle = 90 - angle
            if angle < 0:
                angle += 360
            if angle > 360:
                angle -= 360
            guide["angle"] = round(angle, 2)
            guides.append(guide)
    return guides


def apply_guide_properties(
    guides: List[UfoGuide], properties: GuidePropertyList
) -> None:
    # Update the guides with names and colors from properties
    num_guides = len(guides)
    for prop in properties:
        guide_index = prop["index"]
        assert isinstance(guide_index, int)
        target_guide_index = guide_index - 1  # index is 1-based
        if target_guide_index >= num_guides:
            pp = {k: v for k, v in prop.items() if k != "index"}
            logger.warning(
                "Captain, we lost a guide somewhere. "
                f"Looked for index {target_guide_index}, but it is outside of "
                f"guide array of length {num_guides}:\n{guides}\n"
                f"Properties meant to be applied were:\n{pp}.\nSkipping properties."
            )
            continue

        guide = guides[target_guide_index]
        if "color" in prop:
            color = prop["color"]
            assert isinstance(color, str)
            r = int(color[1:3], 16) / 0xFF
            try:
                g = int(color[3:5], 16) / 0xFF
            except ValueError:
                g = 0
            try:
                b = int(color[5:7], 16) / 0xFF
            except ValueError:
                b = 0
            guide["color"] = f"{r:0.4f},{g:0.4f},{b:0.4f},1"
        if "name" in prop:
            guide["name"] = prop["name"]



from __future__ import annotations

import logging

from typing import Any, Dict, List, Tuple
from vfbLib.parsers import BaseParser, read_encoded_value


logger = logging.getLogger(__name__)


class AnisotropicInterpolationsParser(BaseParser):
    @classmethod
    def _parse(cls) -> List[int]:
        # The graph used for anisotropic interpolation maps, for all axes.
        assert cls.stream is not None
        values = []
        while True:
            axis_values = []
            try:
                n = read_encoded_value(cls.stream)
                axis_values = [
                    (read_encoded_value(cls.stream), read_encoded_value(cls.stream))
                    for _ in range(n)
                ]
                values.append(axis_values)
            except EOFError:
                return values


class AxisMappingsCountParser(BaseParser):
    @classmethod
    def _parse(cls) -> List[int]:
        # An array of axis mapping counts per axis, for all 4 axes whether they exist
        # or not.
        # 0300 0000  0000 0000  0000 0000  0000 0000
        # -> [3, 0, 0, 0]
        assert cls.stream is not None
        return [cls.read_uint32() for _ in range(4)]


class AxisMappingsParser(BaseParser):
    @classmethod
    def _parse(cls) -> List[Tuple[float, float]]:
        # 10 pairs of (user, design) coordinates per axis.
        # Look at "Axis Mappings Count" to find out which mappings are used in each
        # axis.
        # The trailing unused fields may contain junk and must be ignored.
        assert cls.stream is not None
        mappings = []
        for _ in range(cls.stream.getbuffer().nbytes // 16):
            src_tgt = cls.read_doubles(2)
            mappings.append(src_tgt)

        return mappings


class MasterLocationParser(BaseParser):
    @classmethod
    def _parse(cls) -> Tuple[int, Tuple[Any]]:
        # The location on all 4 axes for this master
        assert cls.stream is not None
        # FIXME: Might also be 2 uint16:
        master_index = cls.read_uint32()
        flags = cls.read_doubles(4)
        return master_index, flags


class PrimaryInstancesParser(BaseParser):
    @classmethod
    def _parse(cls) -> List[Dict[str, Any]]:
        assert cls.stream is not None
        stream = cls.stream
        instances = []
        num_instances = read_encoded_value(stream)
        for _ in range(num_instances):
            name_length = read_encoded_value(stream)
            name = stream.read(name_length).decode("cp1252")
            values = [read_encoded_value(stream) / 10000 for _ in range(4)]
            instances.append({"name": name, "values": values})

        return instances



from __future__ import annotations

from io import BufferedWriter, BytesIO
from struct import pack


def write_encoded_value(
    value: int, stream: BufferedWriter | BytesIO, signed=True
) -> None:
    # Encode an integer value and write it to the stream. The "signed" param
    # only applies to the longest representation form.
    if -107 <= value <= 107:
        # 1-byte representation
        encoded = pack(">B", (value + 0x8B))
    elif 107 < value <= 1131:
        # 2-byte representation
        encoded = pack(">H", (value + 0xF694))
    elif -1131 <= value < -107:
        # 2-byte representation, negative values
        encoded = pack(">H", (-value + 0xFA94))
    else:
        # 5-byte representation
        write_value_5(value, stream, signed)
        return

    stream.write(encoded)


def write_value_5(value: int, stream: BufferedWriter | BytesIO, signed=True) -> None:
    # Write an integer value to the stream using the longest encoding (4 bytes plus
    # marker byte).
    fmt = "i" if signed else "I"
    encoded = pack(f">B{fmt}", 0xFF, value)
    stream.write(encoded)



from __future__ import annotations

import logging

from vfbLib.parsers import BaseParser


logger = logging.getLogger(__name__)


class PostScriptInfoParser(BaseParser):
    """
    A parser that reads data as a double-size float.
    """

    @classmethod
    def _parse(cls):
        values = {}
        values["font_matrix"] = cls.read_doubles(6)
        values["force_bold"] = cls.read_int32()
        values["blue_values"] = [cls.read_int32() for _ in range(14)]
        values["other_blues"] = [cls.read_int32() for _ in range(10)]
        values["family_blues"] = [cls.read_int32() for _ in range(14)]
        values["family_other_blues"] = [cls.read_int32() for _ in range(10)]
        values["blue_scale"] = cls.read_double()
        values["blue_shift"] = cls.read_uint32()
        values["blue_fuzz"] = cls.read_uint32()
        values["std_hw"] = cls.read_uint32()
        values["std_vw"] = cls.read_uint32()
        values["stem_snap_h"] = [cls.read_uint32() for _ in range(12)]
        values["stem_snap_v"] = [cls.read_uint32() for _ in range(12)]
        # The bounding box values only get updated during some actions, e.g.
        # going into PS hinting mode
        values["bounding_box"] = dict(
            zip(
                ["xMin", "yMin", "xMax", "yMax"],
                [cls.read_int16() for _ in range(4)],
            )
        )
        values["adv_width_min"] = cls.read_int32()
        values["adv_width_max"] = cls.read_int32()
        values["adv_width_avg"] = cls.read_int32()
        values["ascender"] = cls.read_int32()
        values["descender"] = cls.read_int32()
        values["x_height"] = cls.read_int32()
        values["cap_height"] = cls.read_int32()
        assert cls.stream.read() == b""
        return values



from __future__ import annotations

import logging

from fontTools.misc.textTools import num2binary

# from math import log2
# from struct import unpack
from typing import Any, Dict, List
from vfbLib.parsers import BaseParser, read_encoded_value


logger = logging.getLogger(__name__)


def pprint_bitmap(bitmap, invert=False) -> List[str]:
    print(bitmap)
    w, h = bitmap["size_pixels"]
    data = bitmap["data"]
    col_bytes = w // 8
    rest = w % 8
    if rest > 0:
        col_bytes += 1
    col_bytes = max(col_bytes, 2)
    # print(f"Cols: {w}, bytes per line: {col_bytes}")
    if "bytes" not in data:
        b = []
    else:
        b = data["bytes"]
    gfx = []
    for y in range(0, h * col_bytes, col_bytes):
        col = ""
        for x in range(col_bytes):
            i = y + x
            if i < len(b):
                byte = b[i]
            else:
                byte = 0

            # print(f"Index: {i}, value: {byte}")
            col += num2binary(byte, bits=8).replace("0", "  ").replace("1", "██")
        gfx.append(col[: w * 2])
    if invert:
        gfx.reverse()
    print("\n".join(gfx))
    print(gfx)
    return gfx


class BaseBitmapParser(BaseParser):
    @classmethod
    def parse_bitmap_data(cls, datalen) -> Dict[str, Any]:
        bitmap: Dict[str, Any] = {}
        if datalen < 2:
            logger.error("parse_bitmap_data: Got datalen", datalen)
            raise ValueError

        rest = 2

        if datalen > 2:
            num_bytes = cls.read_uint8()
            bitmap["num_bytes"] = num_bytes
            data = []
            for _ in range(num_bytes):
                data.append(cls.read_uint8())
            bitmap["bytes"] = data
            rest = datalen - num_bytes - 1

        extra = []
        for _ in range(rest):
            extra.append(cls.read_uint8())

        if extra:
            bitmap["extra"] = extra

        return bitmap


class BackgroundBitmapParser(BaseBitmapParser):
    @classmethod
    def _parse(cls) -> Dict[str, Any]:
        s = cls.stream
        bitmap: Dict[str, Any] = {}
        bitmap["origin"] = (read_encoded_value(s), read_encoded_value(s))
        bitmap["size_units"] = (read_encoded_value(s), read_encoded_value(s))
        bitmap["size_pixels"] = (read_encoded_value(s), read_encoded_value(s))
        datalen = read_encoded_value(s)
        bitmap["len"] = datalen
        bitmap["data"] = cls.parse_bitmap_data(datalen)
        # bitmap["preview"] = pprint_bitmap(bitmap)
        assert s.read() == b""
        return bitmap


class GlyphBitmapParser(BaseBitmapParser):
    @classmethod
    def _parse(cls) -> List[Dict[str, Any]]:
        s = cls.stream
        bitmaps: List[Dict[str, Any]] = []
        num_bitmaps = read_encoded_value(s)
        for _ in range(num_bitmaps):
            bitmap: Dict[str, Any] = {}
            bitmap["ppm"] = read_encoded_value(s)
            bitmap["origin"] = (read_encoded_value(s), read_encoded_value(s))
            bitmap["adv"] = (read_encoded_value(s), read_encoded_value(s))
            bitmap["size_pixels"] = (read_encoded_value(s), read_encoded_value(s))
            datalen = read_encoded_value(s)
            bitmap["len"] = datalen
            bitmap["data"] = cls.parse_bitmap_data(datalen)
            # bitmap["preview"] = pprint_bitmap(bitmap, invert=True)
            bitmaps.append(bitmap)
        assert s.read() == b""
        return bitmaps
